"""Overall output specification."""
from typing import TYPE_CHECKING, Dict, Generator, List, Optional, TextIO

import click
from raesl.compile.ast import components, exprs, types
from raesl.compile.ast.comment_storage import DefaultDocStore, DocElement, DocStore
from raesl.compile.ast.exprs import Value, VariableValue
from raesl.compile.ast.nodes import CompoundVarNode, ElementaryVarNode, GroupNode, Node

if TYPE_CHECKING:
    from raesl.compile.ast.exprs import Expression
    from raesl.compile.ast.relations import RelationDefinition
    from raesl.compile.ast.verbs import VerbPreposDef


class Specification:
    """Main class.

    Attributes:
        types: Types of the specification.
        verb_prepos: Verbs and pre-positions.
        rel_defs: Relation definitions.
        comp_defs: Component definitions.
        world: Root component.
    """

    def __init__(self):
        self.types: Dict[str, types.TypeDef] = {}
        self.verb_prepos: List["VerbPreposDef"] = []
        self.rel_defs: List["RelationDefinition"] = []

        self.comp_defs: List[components.ComponentDefinition] = []
        self.world: Optional[components.ComponentDefinition] = None


class _DumpSpec:
    """Class for dumping the specification to an output stream."""

    def __init__(self, output_stream: Optional[TextIO] = None):
        self.stream = click.get_text_stream("stdout") if output_stream is None else output_stream
        self.done_types: Dict[types.BaseType, int] = {}

    def dump(self, spec: Specification):
        """Write the specification symbolically onto the output stream. Intended use is
        verifying content or debugging.
        """
        self.stream.write("\nTYPES:\n")
        if spec.types:
            text_tdefs = [(tdef.name, self._dump_type(tdef.type)) for tdef in spec.types.values()]
            self.stream.write("\n")
            for name, tname in text_tdefs:
                self.stream.write("typedef {}: {}\n".format(name.tok_text, tname))

        self.stream.write("\nVERBS:\n")
        for verbdef in spec.verb_prepos:
            self.stream.write("    {}\n".format(verbdef))
            self._dump_docs(verbdef, "    ")

        self.stream.write("\nRELATION-DEFINITIONS:\n")
        for rel_def in spec.rel_defs:
            self.stream.write("    def {}:\n".format(rel_def.name.tok_text))
            self._dump_docs(rel_def, "    ")
            for param in rel_def.params:
                tnumber = self.done_types[param.type]
                param_type = "type#{}".format(tnumber)
                text = "{} {}: {}".format(param.direction, param.name.tok_text, param_type)
                if param.multi:
                    text = "multi " + text

                self.stream.write("        {}\n".format(text))

        self.stream.write("\nCOMPONENTS:\n")
        if spec.world is not None:
            self.dump_component(spec.world)

        sorted_comp_defs = sorted(spec.comp_defs, key=lambda cd: cd.name_tok.tok_text)
        for comp_def in sorted_comp_defs:
            self.dump_component(comp_def)

    def dump_component(self, compdef):
        """Dump content of the provided component."""
        # Dump component name as header.
        if compdef.name_tok is None:
            self.stream.write("    world:\n")
        else:
            name = compdef.name_tok.tok_text
            self.stream.write("    def component {}:\n".format(name))
        self._dump_docs(compdef, "    ")

        # Dump variables of a component definition.
        leading = "        "
        for vdef in compdef.variables:
            self.stream.write("{}var {}\n{}  ".format(leading, vdef.name_tok.tok_text, leading))
            self._dump_nodes(vdef.node, prefixes=[leading + "  "])

        # Dump parameters of a component definition.
        for pdef in compdef.parameters:
            prop_text = {False: "", True: ": property"}[pdef.is_property]
            text = "param {}{}".format(pdef.name_tok.tok_text, prop_text)
            self.stream.write(leading + text + "\n  " + leading)
            self._dump_nodes(pdef.node, prefixes=[leading + "  "])

        # Dump variable groups of a component definition.
        for vgroup in compdef.var_groups:
            content = ", ".join(varpart.tok_text for varpart in vgroup.variablepart_names)
            text = "vgroup {}: {}".format(vgroup.name_tok.tok_text, content)
            self.stream.write(leading + text + "\n  " + leading)
            self._dump_nodes(vgroup.node, prefixes=[leading + "  "])

        # Dump component instances of a component definition.
        for compinst in compdef.component_instances:
            def_name = compinst.compdef.name_tok.tok_text
            text = "child-component {}: def={}".format(compinst.inst_name_tok.tok_text, def_name)
            self.stream.write("        " + text + "\n")
            self._dump_docs(compinst, "        ")
            self._dump_compinst_params(compinst)

        # Dump relation instances of a component definition.
        for relinst in compdef.relations:
            def_name = relinst.reldef.name.tok_text
            text = "relation-instance {}: def={}".format(relinst.inst_name_tok.tok_text, def_name)
            self.stream.write("        " + text + "\n")
            self._dump_docs(relinst, "        ")
            self._dump_relinst_params(relinst)

        # Dump goals
        leading = "        "
        for goal in compdef.goals:
            name = "{} goal {}:".format(goal.goal_kind, goal.label_tok.tok_text)
            comptext1 = "comp '{}' {} ({} {}) comp '{}'"
            comptext1 = comptext1.format(
                goal.active_comp.inst_name_tok.tok_text,
                goal.doesaux.tok_text,
                goal.verb.tok_text,
                goal.prepos.tok_text,
                goal.passive_comp.inst_name_tok.tok_text,
            )
            self.stream.write(leading + name + "\n")
            self._dump_docs(goal, leading)
            indent = leading + "    "
            self.stream.write(indent + comptext1 + "\n")
            for f in goal.flows:
                self.stream.write(indent + "flow '{}':\n".format(f.name_tok.tok_text))
                self.stream.write(indent + "  ")
                self._dump_nodes(f.flow_node, prefixes=[indent + "  "])

            self._dump_subclauses(goal.sub_clauses)

        # Dump transformations
        for trans in compdef.transforms:
            name = "{} transform {}:".format(trans.transform_kind, trans.label_tok.tok_text)
            comptext = "  {} ({} {})"
            comptext = comptext.format(
                trans.doesaux_tok.tok_text,
                trans.verb_tok.tok_text,
                trans.prepos_tok.tok_text,
            )

            indent = "        "
            self.stream.write(indent + name + "\n")
            self._dump_docs(trans, indent)
            self.stream.write(indent + comptext + "\n")
            for f in trans.in_flows:
                self.stream.write(indent + "  in-flow '{}':\n".format(f.name_tok.tok_text))
                self.stream.write(indent + "  ")
                self._dump_nodes(f.flow_node, prefixes=[indent + "  "])
            for f in trans.out_flows:
                self.stream.write(indent + "  out-flow '{}':\n".format(f.name_tok.tok_text))
                self.stream.write(indent + "  ")
                self._dump_nodes(f.flow_node, prefixes=[indent + "  "])
            self._dump_subclauses(trans.sub_clauses)

        # Dump needs
        for need in compdef.needs:
            if need.subject is None:
                subjdef_text = " UNDEFINED"
            else:
                subjdef_text = ""
            text = "need {}: subject {}{}".format(
                need.label_tok.tok_text, need.subject_tok.tok_text, subjdef_text
            )
            self.stream.write("        " + text + "\n")
            self._dump_docs(need, "        ")

        # Dump designs
        indent = "        "
        for design in compdef.designs:
            name = "{} design {}".format(design.design_kind, design.label_tok.tok_text)
            self.stream.write(indent + name + "\n")
            self._dump_docs(design, indent)
            self._dump_expr(design.expr, indent + "    ")
            self._dump_subclauses(design.sub_clauses)

        # Dump behavior
        for behavior in compdef.behaviors:
            name = "{} behavior {}:".format(behavior.behavior_kind, behavior.name_tok.tok_text)
            self.stream.write("        " + name + "\n")
            self._dump_docs(behavior, "        ")
            for case in behavior.cases:
                self.stream.write("            case " + case.name_tok.tok_text + ":\n")
                self._dump_conds(case.conditions)
                self._dump_results(case.results)
            if behavior.default_results is None:
                self.stream.write("            no default case.\n")
            else:
                self.stream.write("            default:\n")
                self._dump_results(behavior.default_results)

    def _dump_expr(self, expr: "Expression", indent: str):
        first = True
        for ex in unfold_disjunction(expr):
            if first:
                self.stream.write(indent)
                first = False
            else:
                self.stream.write("{}or ".format(indent))

            if isinstance(ex, exprs.RelationComparison):
                # Construct text and node for RHS
                rv = ex.rhs_varval
                if isinstance(rv, Value):
                    rhs_text = "{}[{}]".format(rv.value.tok_text, next(iter(rv.get_units())))
                    rhs_node = None
                else:
                    assert isinstance(rv, VariableValue)
                    rhs_text = rv.var_tok.tok_text
                    rhs_node = rv.var_node

                text = "{} {} {} {}".format(
                    ex.lhs_var.var_tok.tok_text,
                    ex.isaux_tok.tok_text,
                    ex.math_compare,
                    rhs_text,
                )
                self.stream.write(text + "\n")
                self.stream.write(indent)
                self._dump_nodes(ex.lhs_var.var_node, prefixes=[indent])
                if rhs_node is not None:
                    self.stream.write(indent)
                    self._dump_nodes(rhs_node, prefixes=[indent])

            else:
                assert isinstance(ex, exprs.ObjectiveComparison)

                aims = []
                if ex.maximize:
                    aims.append("max")
                if ex.minimize:
                    aims.append("min")
                if not aims:
                    aims.append("UNKNOWN")
                text = "{} {} {}".format(
                    "+".join(aims), ex.aux_tok.tok_text, ex.lhs_var.var_tok.tok_text
                )
                self.stream.write(text + "\n")
                self.stream.write(indent)
                self._dump_nodes(ex.lhs_var.var_node, prefixes=[indent])

    def _dump_nodes(self, node: Node, prefixes: List[str] = None):
        """Recursively dump the tree of var/param nodes rooted at 'node', with
        'prefixes' containing the leading text-fields if not None.
        """
        if prefixes is None:
            prefixes = []

        if isinstance(node, ElementaryVarNode):
            tnumber = self.done_types[node.the_type]
            self.stream.write("- type#{}\n".format(tnumber))
            self._dump_docs(node, "".join(prefixes))
        elif isinstance(node, CompoundVarNode):
            last_index = len(node.child_nodes) - 1
            prefixes.append(None)
            for index, subnode in enumerate(node.child_nodes):
                name = subnode.name_tok.tok_text
                if index > 0:
                    self.stream.write("".join(prefixes[:-1]))

                self.stream.write("\\- " + name + " ")
                if index < last_index:
                    prefixes[-1] = "|" + " " * (3 + len(name))
                else:
                    prefixes[-1] = " " + " " * (3 + len(name))

                self._dump_nodes(subnode, prefixes)
                index = index + 1

            del prefixes[-1]
        elif isinstance(node, GroupNode):
            last_index = len(node.child_nodes) - 1
            prefixes.append(None)
            for index, subnode in enumerate(node.child_nodes):
                name = subnode.name_tok.tok_text
                if index > 0:
                    self.stream.write("".join(prefixes[:-1]))

                self.stream.write("@--")
                if index < last_index:
                    prefixes[-1] = "|  "
                else:
                    prefixes[-1] = "   "

                self._dump_nodes(subnode, prefixes)
                index = index + 1

            del prefixes[-1]

    def _dump_docs(self, element: DocElement, indent_text: str):
        """Dump documentation of the given element if it has any. If no documentation
        has been added, report there is no documentation.

        Arguments:
            element: Element to check for documentation.
            indent_text: Leading text printed at each line by the function.
        """
        comments = element.get_comment()
        if not comments:
            self.stream.write(indent_text + "  #< No doc-comments found.\n")
            return

        comment_text = " ".join(comments)
        self.stream.write(indent_text + "  #< {}\n".format(comment_text))

    def _dump_type(self, btp: types.BaseType) -> str:
        """Dump unique types onto the self.output_stream while assigning them names
        of the form 'type#<num>'. Each type is output once.

        Arugments:
            btp: Type to dump.

        Returns:
            Name of the type during the dump process.
        """
        if btp in self.done_types:
            return "type#{}".format(self.done_types[btp])

        number = len(self.done_types) + 1
        self.done_types[btp] = number
        my_typename = "type#" + str(number)

        if isinstance(btp, types.Compound):
            indent = "    "

            type_names = [self._dump_type(field.type) for field in btp.fields]
            self.stream.write("{} = bundle\n".format(my_typename))
            for field, tname in zip(btp.fields, type_names):
                self.stream.write(indent + "{}: {}\n".format(field.name.tok_text, tname))
            self.stream.write("{} end\n".format(my_typename))
            self.stream.write("\n")

        else:
            assert isinstance(btp, types.ElementaryType)
            if btp.parent is not None:
                parent_name = self._dump_type(btp.parent)
                self.stream.write("{} = type <-- {}\n".format(my_typename, parent_name))
            else:
                self.stream.write("{} = type\n".format(my_typename))

            if btp.units:
                self.stream.write("    units: " + ", ".join(u.tok_text for u in btp.units) + "\n")

            if btp.intervals:
                prefix = "    intervals: "
                for ival in btp.intervals:
                    if ival[0] is None:
                        lbound = "*"
                    else:
                        lbound = str(ival[0])
                    if ival[1] is None:
                        ubound = "*"
                    else:
                        ubound = str(ival[1])

                    self.stream.write(prefix + "(=> {}, <= {})\n".format(lbound, ubound))
                    prefix = " " * len(prefix)

        return my_typename

    def _dump_compinst_params(self, compinst):
        """Dump the arguments of the component instance."""
        if not compinst.arguments:
            self.stream.write("            No instance argument!\n")
            return

        for arg, param in zip(compinst.arguments, compinst.compdef.parameters):
            arg_text = "arg {}".format(arg.name_tok.tok_text)
            param_text = "param {}".format(param.name_tok.tok_text)
            leading = "            "
            self.stream.write("{}{} <-> {}\n".format(leading, arg_text, param_text))
            self.stream.write(leading + "  ")
            self._dump_nodes(arg.argnode, prefixes=[leading + "  "])

    def _dump_relinst_params(self, relinst):
        """Dump the arguments of the relation instance."""
        if not relinst.arguments:
            self.stream.write("            No instance argument!\n")
            return

        for args, param in zip(relinst.arguments, relinst.reldef.params):
            arg_text = [arg.name_tok.tok_text for arg in args]
            param_text = "param('{}', {}, multi={})".format(
                param.name.tok_text, param.direction, param.multi
            )
            leading = "            "
            self.stream.write("{}{} <-> {}\n".format(leading, arg_text, param_text))
            for arg in args:
                self.stream.write(leading + "  ")
                self._dump_nodes(arg.argnode, prefixes=[leading + "  "])

    def _dump_subclauses(self, clauses):
        leading = "            "
        for clause in clauses:
            text = "subclause {}:\n".format(clause.label_tok.tok_text)
            self.stream.write(leading + text)
            self._dump_expr(clause.expr, leading + "  ")

    def _dump_conds(self, conditions):
        leading = "                "
        for cond in conditions:
            self.stream.write("{}{}: if\n".format(leading, cond.name_tok.tok_text))
            self._dump_expr(cond.comparison, indent=leading + "  ")

    def _dump_results(self, results):
        leading = "                "
        for res in results:
            self.stream.write("{}{}: then\n".format(leading, res.name_tok.tok_text))
            self._dump_expr(res.result, indent=leading + "  ")


def unfold_disjunction(expr: "Expression") -> List["Expression"]:
    """Convert the child expressions of top-level disjunction expressions to a list."""
    if isinstance(expr, exprs.Disjunction):
        result = []
        for child in expr.childs:
            result.extend(unfold_disjunction(child))
        return result
    else:
        return [expr]


def dump(spec: Specification, output_stream: Optional[TextIO] = None):
    """Dump the provided specification to an output.

    Arguments:
        spec: Specification to dump.
        output_stream: Stream to write to, None means stdout.
    """
    dump_spec = _DumpSpec(output_stream)
    dump_spec.dump(spec)


def get_doc_comment_spec_elements(
    spec: Specification,
) -> Generator["DocStore", None, None]:
    """Retrieve the specification elements interested in getting documentation comments
    from the input.

    Note: Component definitions are added through components
        get_doc_comment_comp_elements

    Note: The implementation is highly knowledgeable about data structure here with
        respect to the implementation.
    """
    elem: DefaultDocStore
    for elem in spec.verb_prepos:
        assert isinstance(elem, DocStore)
        if elem.doc_tok:
            yield elem

    for elem in spec.rel_defs:
        assert isinstance(elem, DocStore)
        if elem.doc_tok:
            yield elem



"""Line matcher for argument lists.

Note this file is imported from other machine files rather than providing argument list
processing itself.
"""
from typing import TYPE_CHECKING, Dict, List

if TYPE_CHECKING:
    from raesl.compile.scanner import Token

ARGUMENT_LINE_SPEC = """
argument_line:
    start initial;
    start -> s1 [STAR_TK];
    s1 -> s2 [DOTTEDNAME] tag=argument;
    s2 -> s1 [COMMA_TK];

    end accept=argument_line;
    s2 -> end [NL_TK];
    s2 -> end [EOF_TK];
"""


def process_argument_list_line(tags: Dict[str, List["Token"]]) -> List["Token"]:
    """Extract the argument names from the collected tags."""
    return tags["argument"]



"""ESL to Doc Command Line Interface."""
import sys
from typing import Iterable, Optional

import click
import raesl.doc
from raesl import logger

run = raesl.doc.convert


@click.command("doc")
@click.argument("paths", nargs=-1, type=click.Path(exists=True, file_okay=True, dir_okay=True))
@click.option(
    "--output",
    "-o",
    default=raesl.doc.OUTPUT,
    type=click.Path(file_okay=True, dir_okay=False, writable=True),
    help="Output file to write to.",
)
@click.option(
    "--language",
    "-l",
    default=raesl.doc.LANGUAGE,
    type=click.Choice(["en", "nl"], case_sensitive=False),
    help="Output document language.",
)
@click.option(
    "--title",
    "-t",
    default=raesl.doc.TITLE,
    type=click.STRING,
    help="Output document title.",
)
@click.option(
    "--prologue",
    "-p",
    default=raesl.doc.PROLOGUE,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
    help="Optional prologue document to include (Markdown).",
)
@click.option(
    "--epilogue",
    "-e",
    default=raesl.doc.EPILOGUE,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
    help="Optional epilogue document to include (Markdown).",
)
@click.option(
    "--rich",
    "-r",
    default=raesl.doc.RICH,
    type=click.Choice(["tex", "md", "off"], case_sensitive=False),
    help="Format of rich output to use.",
)
@click.option(
    "--force",
    "-f",
    default=raesl.doc.FORCE,
    is_flag=True,
    help="Force overwrite of output file.",
)
@click.option(
    "--dry",
    "-d",
    default=raesl.doc.DRY,
    is_flag=True,
    help="Dry run. Skip creating an output document.",
)
def doc(
    paths: Iterable[str],
    output: str,
    language: str,
    title: str,
    prologue: Optional[str],
    epilogue: Optional[str],
    rich: Optional[str],
    force: bool,
    dry: bool,
):
    """Convert ESL files and/or directories to a formatted document."""
    logger.info("This is the Ratio ESL Doc command line utility.")
    logger.info(f"Populating '{output}', titled as '{title}' in language '{language}'...")
    try:
        run(
            *paths,
            output=output,
            language=language,
            title=title,
            prologue=prologue,
            epilogue=epilogue,
            rich=rich,
            force=force,
            dry=dry,
        )
        logger.info("Doc generation done!")
    except Exception as e:
        logger.error(str(e))
        sys.exit(1)


if __name__ == "__main__":
    doc()



"""State machines to recognize lines of ESL."""
from typing import List, Set

from raesl.compile import machine_files
from raesl.compile.state_machine import StateMachine


def get_all_line_machines() -> List[StateMachine]:
    """Get all line state machines (for testing)."""
    machines: List[StateMachine] = list(machine_files.collect_line_machines().values())
    return machines


def get_line_machine_names() -> Set[str]:
    """Get the name of all line state machines (for testing)."""
    names = set(machine_files.collect_line_machines().keys())
    return names


def get_line_machine(name: str) -> machine_files.builder.ProcessingStateMachine:
    """Retrieve a line matcher state machine by name."""
    machine = machine_files.collect_line_machines()[name]
    return machine


# Specification defining the ESL language. Here, the token to take the
# transition is a successful match with a line machine (that is, recognition of
# one line ESL). Also, the 'epsilon' token exists here, which is a silent
# transition and can always be taken. Its use is to include other states in the
# search for a line match without copying all the transitions from that state.
# In this ESL specification, [epsilon] is used to recognize header lines while
# processing content.
_ESL_SPEC = """
esl:
    start initial, accept=esl_spec;
    start -> types [DEFINE_TYPE_MACHINE];
        types -> types_may_end [NEW_TYPE_MACHINE];
        types -> types_may_end [TYPE_IS_A_TYPE_MACHINE];
        types -> type_bundle [BUNDLE_TYPE_MACHINE];

        type_bundle -> type_bundle_may_end [BUNDLE_FIELD_MACHINE];

        type_bundle_may_end -> type_bundle_may_end [BUNDLE_FIELD_MACHINE];
        type_bundle_may_end -> types_may_end [epsilon];

        types_may_end -> types_may_end [NEW_TYPE_MACHINE];
        types_may_end -> types_may_end [TYPE_IS_A_TYPE_MACHINE];
        types_may_end -> type_bundle [BUNDLE_TYPE_MACHINE];
        types_may_end -> start [epsilon];


    start -> verbs [DEFINE_VERB_MACHINE];
        verbs -> verbs_may_end [VERB_PREPOS_MACHINE];

        verbs_may_end -> verbs_may_end [VERB_PREPOS_MACHINE];
        verbs_may_end -> start [epsilon];

    start -> reldefs [DEFINE_RELATION_MACHINE];
        reldefs -> new_reldef [RELATION_NAME_LINE_MACHINE];

        new_reldef -> reldef_params [RELATION_PARAMETER_HEADER_MACHINE];

        reldef_params -> reldef_params_may_end [RELATION_PARAMETER_LINE_MACHINE];

        reldef_params_may_end -> reldef_params_may_end [RELATION_PARAMETER_LINE_MACHINE];
        reldef_params_may_end -> reldef_params [RELATION_PARAMETER_HEADER_MACHINE];
        reldef_params_may_end -> new_reldef [RELATION_NAME_LINE_MACHINE];
        reldef_params_may_end -> start [epsilon];

    # Includes both normal definitions as well as 'world'.
    start -> component [DEFINE_COMPONENT_MACHINE];
        component -> start [EMPTY_MACHINE];

        component -> variables [VARIABLE_HEADER_MACHINE];
        component -> parameters [PARAMETER_HEADER_MACHINE];
        component -> groups [GROUP_SECTION_HEADER_MACHINE];
        component -> components [COMPONENT_HEADER_MACHINE];
        component -> needs [NEED_HEADER_MACHINE];
        component -> goals [GOAL_HEADER_MACHINE];
        component -> transforms [TRANSFORM_HEADER_MACHINE];
        component -> designs [DESIGN_HEADER_MACHINE];
        component -> relations [RELATION_HEADER_MACHINE];
        component -> behavior [BEHAVIOR_HEADER_MACHINE];
        component -> comments [COMMENT_HEADER_MACHINE];

        component_may_end -> variables [VARIABLE_HEADER_MACHINE];
        component_may_end -> parameters [PARAMETER_HEADER_MACHINE];
        component_may_end -> groups [GROUP_SECTION_HEADER_MACHINE];
        component_may_end -> components [COMPONENT_HEADER_MACHINE];
        component_may_end -> needs [NEED_HEADER_MACHINE];
        component_may_end -> goals [GOAL_HEADER_MACHINE];
        component_may_end -> transforms [TRANSFORM_HEADER_MACHINE];
        component_may_end -> designs [DESIGN_HEADER_MACHINE];
        component_may_end -> relations [RELATION_HEADER_MACHINE];
        component_may_end -> behavior [BEHAVIOR_HEADER_MACHINE];
        component_may_end -> comments [COMMENT_HEADER_MACHINE];


    # Variables
    variables -> variables_may_end [VARIABLE_LINE_MACHINE];

    variables_may_end -> variables_may_end [VARIABLE_LINE_MACHINE];
    variables_may_end -> start [epsilon];
    variables_may_end -> component_may_end [epsilon];


    # Parameters
    parameters -> parameters_may_end [PARAMETER_LINE_MACHINE];

    parameters_may_end -> parameters_may_end [PARAMETER_LINE_MACHINE];
    parameters_may_end -> start [epsilon];
    parameters_may_end -> component_may_end [epsilon];


    # Groups
    groups -> group_started [GROUP_START_MACHINE];

    group_started -> group_may_end [GROUP_ARGUMENT_LINE_MACHINE];

    group_may_end -> group_may_end [GROUP_ARGUMENT_LINE_MACHINE];
    group_may_end -> group_started [GROUP_START_MACHINE];
    group_may_end -> start [epsilon];
    group_may_end -> component_may_end [epsilon];


    # Components
    components -> components_args [COMPONENT_INSTANCE_WITH_ARGS_MACHINE];
    components -> components_may_end [COMPONENT_INSTANCE_NO_ARGS_MACHINE];

    components_args -> components_args_may_end [COMPONENT_ARGUMENT_MACHINE];

    components_args_may_end -> components_args_may_end [COMPONENT_ARGUMENT_MACHINE];
    components_args_may_end -> components_args [COMPONENT_INSTANCE_WITH_ARGS_MACHINE];
    components_args_may_end -> components_may_end [COMPONENT_INSTANCE_NO_ARGS_MACHINE];
    components_args_may_end -> start [epsilon];
    components_args_may_end -> component_may_end [epsilon];

    components_may_end -> components_args [COMPONENT_INSTANCE_WITH_ARGS_MACHINE];
    components_may_end -> components_may_end [COMPONENT_INSTANCE_NO_ARGS_MACHINE];
    components_may_end -> start [epsilon];
    components_may_end -> component_may_end [epsilon];


    # Needs
    needs -> needs_may_end [NEED_LINE_MACHINE];

    needs_may_end -> needs_may_end [NEED_LINE_MACHINE];
    needs_may_end -> component_may_end [epsilon];
    needs_may_end -> start [epsilon];


    # Goals
    goals -> goals_may_end [GOAL_MAIN_NO_SUBS_MACHINE];
    goals -> goals_subs [GOAL_MAIN_WITH_SUBS_MACHINE];

    goals_subs -> goals_subs_may_end [GOAL_SUB_CLAUSE_MACHINE];

    goals_subs_may_end -> goals_subs_may_end [GOAL_SUB_CLAUSE_MACHINE];
    goals_subs_may_end -> goals_may_end [GOAL_MAIN_NO_SUBS_MACHINE];
    goals_subs_may_end -> goals_subs [GOAL_MAIN_WITH_SUBS_MACHINE];
    goals_subs_may_end -> component_may_end [epsilon];
    goals_subs_may_end -> start [epsilon];

    goals_may_end -> goals_may_end [GOAL_MAIN_NO_SUBS_MACHINE];
    goals_may_end -> goals_subs [GOAL_MAIN_WITH_SUBS_MACHINE];
    goals_may_end -> component_may_end [epsilon];
    goals_may_end -> start [epsilon];


    # Transforms
    transforms -> transforms_may_end [TRANSFORM_MAIN_NO_SUBS_MACHINE];
    transforms -> transforms_subs [TRANSFORM_MAIN_WITH_SUBS_MACHINE];

    transforms_subs -> transforms_subs_may_end [TRANSFORM_SUB_CLAUSE_MACHINE];

    transforms_subs_may_end -> transforms_subs_may_end [TRANSFORM_SUB_CLAUSE_MACHINE];
    transforms_subs_may_end -> transforms_may_end [TRANSFORM_MAIN_NO_SUBS_MACHINE];
    transforms_subs_may_end -> transforms_subs [TRANSFORM_MAIN_WITH_SUBS_MACHINE];
    transforms_subs_may_end -> component_may_end [epsilon];
    transforms_subs_may_end -> start [epsilon];

    transforms_may_end -> transforms_may_end [TRANSFORM_MAIN_NO_SUBS_MACHINE];
    transforms_may_end -> transforms_subs [TRANSFORM_MAIN_WITH_SUBS_MACHINE];
    transforms_may_end -> component_may_end [epsilon];
    transforms_may_end -> start [epsilon];


    # Designs
    designs -> designs_may_end [DESIGN_NO_SUBS_MACHINE];
    designs -> designs_subs [DESIGN_WITH_SUBS_MACHINE];

    designs_subs -> designs_subs_may_end [DESIGN_SUB_CLAUSE_MACHINE];

    designs_subs_may_end -> designs_subs_may_end [DESIGN_SUB_CLAUSE_MACHINE];
    designs_subs_may_end -> designs_may_end [DESIGN_NO_SUBS_MACHINE];
    designs_subs_may_end -> designs_subs [DESIGN_WITH_SUBS_MACHINE];
    designs_subs_may_end -> component_may_end [epsilon];
    designs_subs_may_end -> start [epsilon];

    designs_may_end -> designs_may_end [DESIGN_NO_SUBS_MACHINE];
    designs_may_end -> designs_subs [DESIGN_WITH_SUBS_MACHINE];
    designs_may_end -> component_may_end [epsilon];
    designs_may_end -> start [epsilon];


    # Behavior
    behavior -> behavior_entry [BEHAVIOR_NAME_MACHINE];

    behavior_entry -> behavior_case [BEHAVIOR_CASE_MACHINE];

    behavior_case -> behavior_when [BEHAVIOR_WHEN_MACHINE];
    behavior_case -> behavior_otherwise [BEHAVIOR_WHEN_OTHERWISE_MACHINE];

    behavior_when -> behavior_when_cond [BEHAVIOR_WHEN_CONDITION_MACHINE];

    behavior_when_cond -> behavior_when_cond [BEHAVIOR_WHEN_CONDITION_MACHINE];
    behavior_when_cond -> behavior_then [BEHAVIOR_THEN_MACHINE];

    behavior_otherwise -> behavior_then [BEHAVIOR_THEN_MACHINE];

    behavior_then -> behavior_then_result [BEHAVIOR_THEN_RESULT_MACHINE];

    behavior_then_result -> behavior_then_result [BEHAVIOR_THEN_RESULT_MACHINE];
    behavior_then_result -> behavior_case [BEHAVIOR_CASE_MACHINE];
    behavior_then_result -> behavior_entry [BEHAVIOR_NAME_MACHINE];
    behavior_then_result -> component_may_end [epsilon];
    behavior_then_result -> start [epsilon];


    # Relations
    relations -> relations_arguments [RELATION_INSTANCE_MACHINE];

    relations_arguments -> relations_subs [RELATION_ARGUMENT_HEADER_MACHINE];

    relations_subs -> relations_may_end [RELATION_ARGUMENT_MACHINE];

    relations_may_end -> relations_may_end [RELATION_ARGUMENT_MACHINE];
    relations_may_end -> relations_arguments [RELATION_INSTANCE_MACHINE];
    relations_may_end -> relations_subs [RELATION_ARGUMENT_HEADER_MACHINE];
    relations_may_end -> component_may_end [epsilon];
    relations_may_end -> start [epsilon];


    # Comments
    comments -> comments_may_end [COMMENT_LINE_MACHINE];

    comments_may_end -> comments_may_end [COMMENT_LINE_MACHINE];
    comments_may_end -> component_may_end [epsilon];
    comments_may_end -> start [epsilon];
"""  # noqa
_BUILDER = machine_files.builder.StateMachineBuilder()
ESL_MACHINE = _BUILDER.create(_ESL_SPEC)



"""Line matcher state machines for comments."""
from typing import TYPE_CHECKING

from raesl.compile.machine_files import typing

if TYPE_CHECKING:
    from raesl.compile.typechecking.ast_builder import AstBuilder

_COMMENT_HEADER_SPEC = """
comment_header:
    start initial;
    start -> s1 [COMMENT_KW] tag=comment;

    end accept=comment_header;
    s1 -> end [NL_TK];
    s1 -> end [EOF_TK];
"""


def _process_comment_header(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    builder.notify_new_section(tags["comment"][0], False)


_COMMENT_LINE_SPEC = """
comment_line:
    start initial;
    start -> s1 [DOTTEDNAME] tag=name;

    end accept=comment_name;
    s1 -> end [NL_TK];
    s1 -> end [EOF_TK];
"""


def _process_comment_name(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.add_comment(tags["name"][0])


MACHINES: typing.MachineTripletList = [
    ("COMMENT_HEADER_MACHINE", _COMMENT_HEADER_SPEC, _process_comment_header),
    ("COMMENT_LINE_MACHINE", _COMMENT_LINE_SPEC, _process_comment_name),
]



"""ESL Language Server config object."""
from pathlib import Path
from typing import Any, List, Optional

from pygls.workspace import Workspace
from raesl import logger
from raesl.utils import check_output_path, cleanup_path, get_esl_paths


class EslConfig:
    """ESL workspace configuration.

    Attributes:
        paths: ESL input paths derived from workspace config.
        output: Graph output path derived from workspace config.
    """

    def __init__(self):
        self.paths: Optional[List[Path]] = None
        self.output: Optional[Path] = None

    def get_paths(
        self,
        ws: Workspace,
        doc_uri: Optional[str] = None,
    ) -> Optional[List[Path]]:
        """Current set of ESL paths."""
        if doc_uri is None:
            return self.paths

        doc_path = cleanup_path(doc_uri)

        # If no paths are set or if outside of current scope: return doc_path.
        # Otherwise, we have paths set and the doc_path is part of them.
        if self.paths is None:
            logger.info("No paths set, returning document path...")
            return [doc_path]
        elif doc_path not in self.paths:
            logger.info("Document path outside current set paths, returning that...")
            return [doc_path]
        else:
            logger.info("Returning set paths...")
            return self.paths

    def parse_config(self, ws: Workspace, cfg: Any):
        """Parse workspace config."""
        logger.info(f"Parsing config: '{cfg}'...")

        # Handle paths variable.
        paths = getattr(cfg, "eslPaths", False)
        if ws and ws.root_path:
            root = ws.root_path
            logger.info(f"Found root '{root}'.")
        else:
            root = None

        if paths:
            logger.info(f"Parsing config paths {paths}...")
            self.paths = get_esl_paths([cleanup_path(p) for p in paths], root=root)
            logger.info(f"Parsed into '{self.paths}'.")

        elif ws is not None and ws.root_path:
            logger.info("Parsing workspace root...")
            self.paths = get_esl_paths([ws.root_path], root=root)

        else:
            logger.info("No paths found so far, setting None.")
            self.paths = None

        # Handle graph output variable.
        try:
            output = getattr(cfg, "graphPath", None)
            self.output = check_output_path(cleanup_path(output), True, root=root)
        except (ValueError, TypeError):
            self.output = None
        logger.info(f"Set graph output to '{self.output}'.")

        logger.info("Parsed config.")



"""Lexer for on-demand recognition of tokens.

As the language has unrestrained text in its needs, lexing beforehand is not
going to work in all cases. Instead, the scanner tries to match tokens on demand.

Also there is overlap in matching between tokens (a NONSPACE expression matches
almost all other tokens as well, and a NAME expression matches all keywords).
The rule applied here (by means of sorting edges in the state machines) is that
specific wins from generic. For example, if at some point both the OR_KW and
the NONSPACE token may be used, and the text is "or", the OR_KW token is chosen.
"""
import re
from functools import partial
from typing import Dict, List, Optional, Pattern, Tuple

from raesl import logger, utils
from raesl.types import Location

# Whitespace matching RE.
SPACE_RE = re.compile("[ \\t\\r]+")

# A letter, followed by letters, digits or dashes, where it may not end with a dash.
_FIRST_NAME_PAT = "[A-Za-z][A-Za-z0-9]*(?:[-_][A-Za-z0-9]+)*"
# Like _FIRST_NAME_PAT, but allow both letters and digits as first character.
_OTHER_NAME_PAT = "[A-Za-z0-9]+(?:[-_][A-Za-z0-9]+)*"
# A _FIRST_NAME_PAT followed by zero or more _OTHER_NAME_PAT, separated by dots.
_DOTTED_NAME_PAT = _FIRST_NAME_PAT + "(?:\\." + _OTHER_NAME_PAT + ")*"
# Characters that should not appear after most tokens (they should be part of them).
_KWORD_AVOID_PAT = "(?![-_A-Za-z0-9])"
# Characters that should not appear after names (it is less restrictive than
# _KWORD_AVOID_PAT).
_NAME_AVOID_PAT = "(?![A-Za-z0-9])"


# Helper method to reduce clutter.
comp = partial(re.compile, flags=re.IGNORECASE)

TOKENS: Dict[str, Pattern] = {
    "A_KW": comp("an?" + _KWORD_AVOID_PAT),
    "AND_KW": comp("and" + _KWORD_AVOID_PAT),
    "APPLIES_KW": comp("applies" + _KWORD_AVOID_PAT),
    "APPROXIMATELY_KW": comp("approximately" + _KWORD_AVOID_PAT),
    "ARGUMENT_KW": comp("arguments?" + _KWORD_AVOID_PAT),
    "AT_KW": comp("at" + _KWORD_AVOID_PAT),
    "BE_KW": comp("be" + _KWORD_AVOID_PAT),
    "BEHAVIOR_CONSTRAINT_KW": comp("behavior-constraints?" + _KWORD_AVOID_PAT),
    "BEHAVIOR_REQUIREMENT_KW": comp("behavior-requirements?" + _KWORD_AVOID_PAT),
    "BUNDLE_KW": comp("bundle" + _KWORD_AVOID_PAT),
    "CASE_KW": comp("case" + _KWORD_AVOID_PAT),
    "COLON_TK": comp(":"),
    "COMMA_TK": comp(","),
    "COMMENT_KW": comp("comments?" + _KWORD_AVOID_PAT),
    "COMPONENT_KW": comp("components?" + _KWORD_AVOID_PAT),
    "COULD_KW": comp("could" + _KWORD_AVOID_PAT),
    "DEFINE_KW": comp("define" + _KWORD_AVOID_PAT),
    "DESIGN_CONSTRAINT_KW": comp("design-constraints?" + _KWORD_AVOID_PAT),
    "DESIGN_REQUIREMENT_KW": comp("design-requirements?" + _KWORD_AVOID_PAT),
    "DOES_KW": comp("does" + _KWORD_AVOID_PAT),
    "EMPTY_KW": comp("empty" + _KWORD_AVOID_PAT),
    "ENUMERATION_KW": comp("enumeration" + _KWORD_AVOID_PAT),
    "EQUAL_KW": comp("equal" + _KWORD_AVOID_PAT),
    "GOAL_CONSTRAINT_KW": comp("goal-constraints?" + _KWORD_AVOID_PAT),
    "GOAL_REQUIREMENT_KW": comp("goal-requirements?" + _KWORD_AVOID_PAT),
    "GREATER_KW": comp("greater" + _KWORD_AVOID_PAT),
    "GROUP_KW": comp("group" + _KWORD_AVOID_PAT),
    "IS_KW": comp("(?:is|are)" + _KWORD_AVOID_PAT),
    "LEAST_KW": comp("least" + _KWORD_AVOID_PAT),
    "MAXIMIZED_KW": comp("maximized" + _KWORD_AVOID_PAT),
    "MINIMIZED_KW": comp("minimized" + _KWORD_AVOID_PAT),
    "MORE_KW": comp("more" + _KWORD_AVOID_PAT),
    "MOST_KW": comp("most" + _KWORD_AVOID_PAT),
    "MUST_KW": comp("must" + _KWORD_AVOID_PAT),
    "NEED_KW": comp("needs?" + _KWORD_AVOID_PAT),
    "NO_KW": comp("no" + _KWORD_AVOID_PAT),
    "NOT_KW": comp("not" + _KWORD_AVOID_PAT),
    "OF_KW": comp("of" + _KWORD_AVOID_PAT),
    "ONE_KW": comp("one" + _KWORD_AVOID_PAT),
    "OR_KW": comp("or" + _KWORD_AVOID_PAT),
    "OTHER_KW": comp("other" + _KWORD_AVOID_PAT),
    "PARAMETER_KW": comp("parameters?" + _KWORD_AVOID_PAT),
    "PROPERTY_KW": comp("property" + _KWORD_AVOID_PAT),
    "RELATION_KW": comp("relations?" + _KWORD_AVOID_PAT),
    "RELATING_KW": comp("relating" + _KWORD_AVOID_PAT),
    "REQUIRING_KW": comp("requiring" + _KWORD_AVOID_PAT),
    "RETURNING_KW": comp("returning" + _KWORD_AVOID_PAT),
    "SHALL_KW": comp("shall" + _KWORD_AVOID_PAT),
    "SHOULD_KW": comp("should" + _KWORD_AVOID_PAT),
    "SMALLER_KW": comp("smaller" + _KWORD_AVOID_PAT),
    "STAR_TK": comp("\\*"),
    "SUB_CLAUSES_KW": comp("subclauses?" + _KWORD_AVOID_PAT),
    "THAN_KW": comp("than" + _KWORD_AVOID_PAT),
    "THEN_KW": comp("then" + _KWORD_AVOID_PAT),
    "TO_KW": comp("to" + _KWORD_AVOID_PAT),
    "TRANSFORM_CONSTRAINT_KW": comp("transformation-constraints?" + _KWORD_AVOID_PAT),
    "TRANSFORM_REQUIREMENT_KW": comp("transformation-requirements?" + _KWORD_AVOID_PAT),
    "TYPE_KW": comp("types?" + _KWORD_AVOID_PAT),
    "UNIT_KW": comp("units?" + _KWORD_AVOID_PAT),
    "VARIABLE_KW": comp("variables?" + _KWORD_AVOID_PAT),
    "VARIABLE_GROUP_KW": comp("variable-groups?" + _KWORD_AVOID_PAT),
    "VERB_KW": comp("verbs?" + _KWORD_AVOID_PAT),
    "WHEN_KW": comp("when" + _KWORD_AVOID_PAT),
    "WITH_KW": comp("with" + _KWORD_AVOID_PAT),
    "WONT_KW": comp("won'?t" + _KWORD_AVOID_PAT),
    "WORLD_KW": comp("world" + _KWORD_AVOID_PAT),
    "NAME": comp(_FIRST_NAME_PAT + _NAME_AVOID_PAT),
    "DOTTEDNAME": comp(_DOTTED_NAME_PAT + _NAME_AVOID_PAT),
    "NONCOMMA": comp("[^, \\t\\r\\n]+(?:,+[^, \\t\\r\\n]+)*"),
    "NONSPACE": comp("[^ \\t\\r\\n]+"),
    # NL_TK is a special case
    # EOF_TK is a special case
    # DOC_COMMENT_TK is used to store documentation comments internally.
}


def get_token_priority(tok_type: str) -> int:
    """Priority of the tokens. Higher value is less specific.

    Arguments:
        tok_type: Name of the token type.

    Returns:
        Priority of the token.
    """
    tok_prios = {"NONSPACE": 4, "NONCOMMA": 3, "NAME": 2, "DOTTEDNAME": 2, "epsilon": 0}
    return tok_prios.get(tok_type, 1)  # Default priority is 1


class Token:
    """Data of a matched token.

    Arguments:
        tok_type: Type name of the token.
        tok_text: Text of the token.
        fname: Name of the file containing the text.
        offset: Offset of the current position in the input text.
        line_offset: Offset of the first character of the current line in the input
            text.
        line_num: Line number of the current line.
    """

    def __init__(
        self,
        tok_type: str,
        tok_text: str,
        fname: Optional[str],
        offset: int,
        line_offset: int,
        line_num: int,
    ):
        self.tok_type = tok_type
        self.tok_text = tok_text
        self.fname = fname
        self.offset = offset
        self.line_offset = line_offset
        self.line_num = line_num

    def get_location(self, offset: int = 0) -> Location:
        """Get this token's Location."""
        fname = self.fname if self.fname is not None else "unknown-file"

        if offset < 0:
            offset = 0
        elif offset >= len(self.tok_text):
            offset = len(self.tok_text)

        line, col = self.line_num, self.offset - self.line_offset
        return utils.get_location(
            uri=fname,
            start_line=line,
            start_character=col,
            end_line=line,
            end_character=col + offset,
        )

    def __str__(self) -> str:
        return 'Token["{}", {}]'.format(self.tok_type, repr(self.tok_text))


class Lexer:
    """On-demand scanner.

    For debugging token matching, enable the PARSER_DEBUG flag near the top
    of the file. That also enables debug output in the parser.parse_line to
    understand what line is being tried, and which line match steppers are
    running.

    Arguments;
        fname: Name of the file containing the text, may be None.
        text: Input text.
        length: Length of the text.
        offset: Offset of the current position in the text.
        line_offset: Offset of the first character of the current line in the input
            text.
        line_num: Line number of the current line.
        doc_comments: Documentation comments found so far, shared between all scanners.
    """

    def __init__(
        self,
        fname: Optional[str],
        text: str,
        offset: int,
        line_offset: int,
        line_num: int,
        doc_comments: List[Token],
    ):
        self.fname = fname
        self.text = text
        self.length = len(text)
        self.offset = offset
        self.line_offset = line_offset
        self.line_num = line_num
        self.doc_comments = doc_comments

    def copy(self) -> "Lexer":
        """Make copy of self. New scanner at the same position as self."""
        return Lexer(
            self.fname,
            self.text,
            self.offset,
            self.line_offset,
            self.line_num,
            self.doc_comments,
        )

    def get_location(self) -> Location:
        """Get location information of the next token. Note that such a position may be
        at an unexpected place since new-lines are significant. For example, it may be
        at the end of a comment.

        Returns:
            Location information of the next token.
        """
        fname = self.fname if self.fname is not None else "unknown-file"
        line, col = self.get_linecol()
        return utils.get_location(uri=fname, start_line=line, start_character=col)

    def get_linecol(self) -> Tuple[int, int]:
        """Get line and column information of the next token. Note that as new-lines
        are significant, such a position may be at an unexpected place, for example at
        the end of a comment.

        Returns:
            Line and column information of the next token.
        """
        return self.line_num, self.offset - self.line_offset

    def find(self, tok_type: str) -> Optional[Token]:
        """Try to find the requested token.

        Arguments:
            tok_type: Type name of the token.

        Returns:
            Found token, or None.
        """
        self.skip_white()
        pat = TOKENS.get(tok_type)
        if pat:
            match = pat.match(self.text, self.offset)
            if not match:
                logger.debug("Lexer failed {}".format(tok_type))
                return None

            tok = Token(
                tok_type,
                match[0],
                self.fname,
                self.offset,
                self.line_offset,
                self.line_num,
            )
            self.offset = match.end()
            logger.debug("Lexer matched {}".format(tok))
            return tok

        elif tok_type == "NL_TK":
            if self.offset < self.length and self.text[self.offset] == "\n":
                tok = Token(
                    tok_type,
                    "\n",
                    self.fname,
                    self.offset,
                    self.line_offset,
                    self.line_num,
                )
                self.offset = self.offset + 1
                self.line_offset = self.offset
                self.line_num = self.line_num + 1
                logger.debug("Lexer matched {}".format(tok))
                return tok

            logger.debug("Lexer failed {}".format(tok_type))
            return None

        else:
            assert tok_type == "EOF_TK", "Found unrecognized token {}.".format(tok_type)
            if self.offset >= self.length:
                tok = Token(
                    tok_type,
                    "",
                    self.fname,
                    self.length,
                    self.line_offset,
                    self.line_num,
                )
                logger.debug("Lexer matched {}".format(tok))
                return tok

            logger.debug("Lexer failed {}".format(tok_type))
            return None

    def skip_white(self):
        """Skip white space, triple dots, newlines, and comments. Implements the
        following Graphviz diagram:

        digraph white {
            1 -> 1 [label="spc+"]
            1 -> 99 [label="eof"]
            1 -> 4 [label="#.*"]
            1 -> 5 [label="..."]
            1 -> 99 [label="other"]

            4 -> 99 [label="eof"]
            4 -> 99 [label="nl"]

            5 -> 5 [label="spc+"]
            5 -> 99 [label="eof"]
            5 -> 1 [label="nl"]
            5 -> 6 [label="#.*"]
            5 -> REV [label="..."]
            5 -> REV [label="other"]

            6 -> 99 [label="eof"]
            6 -> 1 [label="nl"]
        }

        Jump to non-99 location eats the recognized text, REV means the
        last found "..." was a false positive and must be reverted to just
        before that position.

        Note that \n is a significant token, so it is not skipped everywhere.
        """
        while True:
            # 1:
            match = SPACE_RE.match(self.text, self.offset)
            if match:
                self.offset = match.end(0)

            if self.offset >= self.length:
                return

            if self.text[self.offset] == "#":
                # 4, starting with matching ".*":
                i = self.text.find("\n", self.offset + 1)
                self._save_doc_comment(self.offset, i)
                if i < 0:
                    self.offset = self.length
                    return
                else:
                    self.offset = i  # A '\n' is needed to end the current line in the parser!
                    return

            if self.text.startswith("...", self.offset):
                # 5:
                # Switch to using 'tmp_offset' as the offset, as we may have
                # to revert skipping ... .
                tmp_offset = self.offset + 3
                match = SPACE_RE.match(self.text, tmp_offset)
                if match:
                    tmp_offset = match.end(0)

                if tmp_offset >= self.length:
                    self.offset = self.length
                    return

                char = self.text[tmp_offset]
                if char == "\n":  # "... \n" found
                    self.offset = tmp_offset + 1
                    self.line_offset = self.offset
                    self.line_num = self.line_num + 1
                    continue

                if char == "#":  # "... #.*\n" found?
                    # 6, starting with matching ".*":
                    i = self.text.find("\n", tmp_offset + 1)
                    self._save_doc_comment(tmp_offset, i)
                    if i < 0:
                        self.offset = self.length
                        return
                    else:
                        self.offset = i + 1
                        self.line_offset = self.offset
                        self.line_num = self.line_num + 1
                        continue

                # Continuation of 5.
                # Found more text, '...' was a false positive, don't skip it.
                return

            # Continuation of 1, 'other' case
            return

    def _save_doc_comment(self, hash_char_offset: int, nl_offset: int):
        """Inspect a comment and rescue doc-comments.

        Arguments:
            hash_char_offset: Offset of the '#' character in the text.
            nl_offset: Offset of the next '\n' in the text, negative value means
                no '\n' was found.
        """
        if nl_offset < 0:
            nl_offset = self.length

        if hash_char_offset + 1 >= nl_offset:
            # Was '#\n' or '#<EOF>', definitely not a '#<'
            return

        if self.text[hash_char_offset + 1] != "<":
            return  # Normal comment, don't save.

        if self.doc_comments and self.doc_comments[-1].offset >= hash_char_offset:
            # Several scanners may run in parallel, so it is feasible that the same
            # doc comment is found several times. Skip this comment if it has been
            # seen before.
            return

        comment_text = self.text[hash_char_offset:nl_offset]
        if not comment_text[2:].strip():
            return  # '#<' was empty or just white-space.

        doc_tok = Token(
            "DOC_COMMENT_TK",
            comment_text,
            self.fname,
            hash_char_offset,
            self.line_offset,
            self.line_num,
        )
        self.doc_comments.append(doc_tok)



"""Relation instance type-checking in a component definition."""
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Union, cast

from raesl.compile import diagnostics
from raesl.compile.ast import relations
from raesl.compile.ast.components import InstanceArgument, RelationInstance
from raesl.compile.typechecking import utils
from raesl.compile.typechecking.type_checker import check_type

if TYPE_CHECKING:
    from raesl.compile.ast.components import (
        ComponentDefinition,
        VariableGroup,
        VarParam,
    )
    from raesl.compile.ast.specification import Specification
    from raesl.compile.ast.types import BaseType
    from raesl.compile.scanner import Token
    from raesl.compile.typechecking.compdef_builder import CompDefChildBuilders


class RelInstArgsBlock:
    """A 'block' of arguments of a relation instance. The kind of arguments, and a list
    of argument lines.

    Arguments:
        argkind_tok: Token indicating the kind of arguments specified in the block.
        arg_name_toks: Arguments of the block.
    """

    def __init__(self, argkind_tok: "Token", arg_name_toks: Optional[List["Token"]] = None):
        self.argkind_tok = argkind_tok

        self.arg_name_toks: List["Token"]
        if arg_name_toks is None:
            self.arg_name_toks = []
        else:
            self.arg_name_toks = arg_name_toks


class RelInst:
    """Relation instance while collecting.

    Arguments:
        inst_name_tok: Instance name.
        def_name_tok: Definition name.
        arg_blocks: Blocks with arguments.
    """

    def __init__(
        self,
        inst_name_tok: "Token",
        def_name_tok: "Token",
        arg_blocks: Optional[List[RelInstArgsBlock]] = None,
    ):
        self.inst_name_tok = inst_name_tok
        self.def_name_tok = def_name_tok

        self.arg_blocks: List[RelInstArgsBlock]
        if arg_blocks is None:
            self.arg_blocks = []
        else:
            self.arg_blocks = arg_blocks


class RelInstArgGroup:
    """Temporary data storage of parameters and arguments of a single kind."""

    def __init__(
        self,
        argkind: str,
        parameters: List[relations.RelationDefParameter],
        arguments: List["Token"],
    ):
        self.argkind = argkind
        self.parameters = parameters
        self.arguments = arguments


class CompDefRelInstBuilder:
    """Collect and check relation instances in a component definition.

    Arguments:
        comp_child_builders: Storage of child builders for a component definition.

    Attributes:
        relinsts: Collected relation instances in the component.
        last_relinst: Link to last added instance to allow adding instance arguments.
    """

    def __init__(self, comp_child_builders: "CompDefChildBuilders"):
        self.diag_store = comp_child_builders.diag_store
        self.relinsts: List[RelInst] = []
        self.comp_child_builders = comp_child_builders
        self.last_relinst: Optional[RelInst] = None

    def new_relinst(self, inst_name_tok: "Token", def_name_tok: "Token"):
        """Parser found a new relation instance, store it for future extending by
        the parser, and eventual type checking and adding to the surrounding
        component definition.
        """
        relinst = RelInst(inst_name_tok, def_name_tok)
        self.relinsts.append(relinst)
        self.last_relinst = relinst

    def relinst_argheader(self, argkind_tok: "Token"):
        """Parser found a new direction block for a relation instance, collect it."""
        assert self.last_relinst is not None

        arg_block = RelInstArgsBlock(argkind_tok)
        self.last_relinst.arg_blocks.append(arg_block)

    def add_relinst_arguments(self, name_toks: List["Token"]):
        """Parser found an argument of a direction-block in a relation instance, store
        it for future checking.
        """
        assert self.last_relinst is not None

        self.last_relinst.arg_blocks[-1].arg_name_toks.extend(name_toks)

    def finish_comp(self, comp_def: "ComponentDefinition", spec: "Specification"):
        """Check the collected relation instances, report errors, and add the
        instances to the given component.

        Arguments:
            comp_def: Component definition to extend with the found relation instances.
                Also a source of available variables, parameters, and variable groups.
            spec: Specification being constructed. Source for types and relation
                definitions processed previously.
        """
        self._add_relation_instance_names()

        # Available names of variables, parameters, and variable-groups in the
        # component.
        avail_vps = utils.construct_var_param_map(comp_def)
        avail_vgroups = utils.construct_vargroup_map(comp_def)

        reported_names: Set[str] = set()  # Names already reported to avoid reporting them again.

        # Make access to relation definitions in the specification easy.
        reldefs = dict((reldef.name.tok_text, reldef) for reldef in spec.rel_defs)

        # Check instances.
        for relinst in self.relinsts:
            # Look for a definition.
            reldef = reldefs.get(relinst.def_name_tok.tok_text)
            if reldef is None:
                # Truly undefined relation definition.
                loc = relinst.def_name_tok.get_location()
                name = relinst.def_name_tok.tok_text
                self.diag_store.add(
                    diagnostics.E203("relation definition", name=name, location=loc)
                )
                continue

            def_parameters = CompDefRelInstBuilder._split_definition_parameters_by_kind(reldef)
            instkind_groups: Optional[List[RelInstArgGroup]]
            instkind_groups = self._split_instance_arguments_by_kind(
                relinst, reldef, def_parameters
            )
            if instkind_groups is None:
                continue

            reldef_param_indices = dict((rd_param, i) for i, rd_param in enumerate(reldef.params))
            instance_arguments: List[Optional[List[InstanceArgument]]]
            instance_arguments = [None] * len(
                reldef.params
            )  # Gets filled using 'reldef_param_indices'

            found_error = False
            for group in instkind_groups:
                argkind = group.argkind
                params = group.parameters
                argument_lists = self._group_param_args(
                    argkind, params, group.arguments, relinst, reldef
                )
                if argument_lists is None:
                    found_error = True
                    continue

                # Perform type checking.
                assert len(params) == len(argument_lists)
                for param, args in zip(params, argument_lists):
                    relinst_argument = []  # list due to the 'one or more' feature.

                    # - For singular-value parameters, args is a list of length 1 and
                    #   each value in it must fit in the parameter type.
                    # - For plural-value parameters, args may be longer and
                    #   each value in it must fit in the parameter type.
                    param_input: Tuple[
                        Optional["Token"],
                        Union["BaseType", "VarParam", "VariableGroup"],
                    ]
                    assert param.type is not None
                    param_input = (param.name, param.type)
                    for arg in args:
                        # Convert token of the argument to a node.
                        node = utils.resolve_var_param_group_node(
                            arg,
                            avail_vps,
                            avail_vgroups,
                            reported_names,
                            self.diag_store,
                        )
                        if node is None:
                            found_error = True
                            continue

                        # Check type
                        arg_input = (arg, node)
                        type_problem: Optional[diagnostics.EslDiagnostic]
                        type_problem = check_type(
                            subtype=arg_input,
                            supertype=param_input,
                            allow_subtype_limits=(argkind == "requiring"),
                        )
                        if type_problem is not None:
                            self.diag_store.add(type_problem)
                            found_error = True
                            continue

                        relinst_argument.append(InstanceArgument(arg, node))

                    instance_arguments[reldef_param_indices[param]] = relinst_argument

            if not found_error:
                assert all(ia is not None for ia in instance_arguments)
                instance = RelationInstance(
                    relinst.inst_name_tok,
                    relinst.def_name_tok,
                    cast(List[List[InstanceArgument]], instance_arguments),
                    reldef,
                )
                comp_def.relations.append(instance)

    def _add_relation_instance_names(self):
        """Check whether the relation instances all have unique instance names, else
        report an error.
        """
        elements_by_label: Dict[str, List[Any]] = self.comp_child_builders.elements_by_label
        for relinst in self.relinsts:
            elements_by_label[relinst.inst_name_tok.tok_text].append(relinst)

    @staticmethod
    def _split_definition_parameters_by_kind(
        reldef: relations.RelationDefinition,
    ) -> Dict[str, List[relations.RelationDefParameter]]:
        """Split the parameters of the definition in 3 groups, 'requiring', 'returning',
        and 'relating'.

        Arguments:
            reldef: Relation definition to use.

        Returns:
            Mapping of the parameters in the mentioned 3 groups.
        """
        def_parameters: Dict[str, List[relations.RelationDefParameter]]
        def_parameters = {"requiring": [], "returning": [], "relating": []}
        for param in reldef.params:
            if param.direction == relations.INPUT:
                def_parameters["requiring"].append(param)
            elif param.direction == relations.OUTPUT:
                def_parameters["returning"].append(param)
            else:
                assert param.direction == relations.INPOUT
                def_parameters["relating"].append(param)

        return def_parameters

    def _split_instance_arguments_by_kind(
        self,
        relinst: RelInst,
        reldef: relations.RelationDefinition,
        def_parameters: Dict[str, List[relations.RelationDefParameter]],
    ) -> Optional[List[RelInstArgGroup]]:
        """Split the argument blocks, and check whether instance and definition agree
        with each other. Report errors if discrepancies are found, and merge argument
        blocks.

        Arguments
            relinst: Relation instance to check.
            def_parameters: Parameters of the definition grouped by argument kind.

        Returns:
            Triplets (kind, parameter-defs, argument tokens) wrapped in RelInstArgGroup
                instances, or None if a fatal error was found.
        """
        # Split argument blocks by argument kind.
        inst_blocks: Dict[str, List[RelInstArgsBlock]]
        inst_blocks = {"requiring": [], "returning": [], "relating": []}
        for arg_block in relinst.arg_blocks:
            inst_blocks[arg_block.argkind_tok.tok_text].append(arg_block)

        # Verify blocks for each kind, instance should have one block if definition has
        # one, else it should have no block.
        for argkind, arg_blocks in inst_blocks.items():
            if not def_parameters[argkind]:
                # Definition has no arguments for the kind.
                if arg_blocks:
                    rel_loc = reldef.name.get_location()
                    block_locs = [arg_block.argkind_tok.get_location() for arg_block in arg_blocks]
                    self.diag_store.add(
                        diagnostics.E206(
                            reldef.name.tok_text,
                            argkind,
                            location=rel_loc,
                            blocks=block_locs,
                        )
                    )
                continue  # Ignore any existing arguments.

            # Definition has arguments for the kind.
            if not arg_blocks:
                # But the instance has not.
                inst_loc = relinst.inst_name_tok.get_location()
                def_loc = reldef.name.get_location()
                name = relinst.inst_name_tok.tok_text
                self.diag_store.add(
                    diagnostics.E207(name, argkind, location=inst_loc, definition=def_loc)
                )
                return None  # Fatal error.

            if len(arg_blocks) > 1:
                locs = [arg_block.argkind_tok.get_location() for arg_block in arg_blocks]
                diag = diagnostics.E200(argkind, "parameter block", location=locs[0], dupes=locs)
                diag.severity = diagnostics.WARN
                self.diag_store.add(diag)
                # Not fatal, we merge the blocks below.

        groups = []
        for argkind, arg_blocks in inst_blocks.items():
            parameters = def_parameters[argkind]
            if not parameters:
                continue  # Ignore arguments if there are no parameters for them.

            arguments: List["Token"] = []
            for arg_block in arg_blocks:
                arguments.extend(arg_block.arg_name_toks)

            groups.append(RelInstArgGroup(argkind, parameters, arguments))

        return groups

    def _group_param_args(
        self,
        argkind: str,
        parameters: List[relations.RelationDefParameter],
        arguments: List["Token"],
        relinst: RelInst,
        reldef: relations.RelationDefinition,
    ) -> Optional[List[List["Token"]]]:
        """Group the arguments into pieces such that each parameter has exactly one
        piece. Mostly this means that the first multi-value parameter takes up all
        slack.

        Arguments:
            argkind: Kind of parameters, one of ('requiring', 'returning', 'relating')
            parameters: Parameters of the definition for the argkind.
            arguments: Arguments of the instance for the argkind.
            relinst: Relation instance.
            reldef: Relation definition.

        Returns:
            None if a fatal error occurred, else grouped arguments.
        """
        # Are there multi-value parameters?
        multiple_index = None
        for i, param in enumerate(parameters):
            if param.multi:
                multiple_index = i
                break

        # Verify length of both lists, and split the argument list such that
        # each parameter gets a list of values.
        params_length = len(parameters)
        args_length = len(arguments)

        if multiple_index is None:
            # All single value parameters.
            if params_length != args_length:
                # Unequal number of arguments and parameters, report an error.
                self.diag_store.add(
                    diagnostics.E221(
                        "argument",
                        args_length,
                        params_length,
                        location=relinst.inst_name_tok.get_location(),
                        references=[reldef.name.get_location()],
                    )
                )
                return None

        else:
            if args_length < params_length:
                name = relinst.inst_name_tok.tok_text
                inst_loc = relinst.inst_name_tok.get_location()
                def_loc = reldef.name.get_location()
                self.diag_store.add(
                    diagnostics.E208(
                        name,
                        argkind,
                        params_length - args_length,
                        location=inst_loc,
                        definition=def_loc,
                    )
                )
                return None

        splitted = utils.split_arguments(params_length, multiple_index, arguments)
        return splitted



"""Code for collecting and adding goals to component definitions."""
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from raesl.compile.ast import components
from raesl.compile.typechecking import utils
from raesl.compile.typechecking.expr_checker import ExprChecker
from raesl.compile.typechecking.goal_transform_base import GoalTransformBaseBuilder

if TYPE_CHECKING:
    from raesl.compile.ast.specification import Specification
    from raesl.compile.scanner import Token
    from raesl.compile.typechecking.compdef_builder import CompDefChildBuilders


class CompDefGoalBuilder(GoalTransformBaseBuilder):
    """Collect goals of a component from the parser, check them, and eventually add them
    to the surrounding component definition.

    Arguments:
        goals: Found goals in the component.
        goal_kind: Last found kind of goal section, 'requirement' or 'constraint'.
    """

    def __init__(self, comp_child_builders: "CompDefChildBuilders"):
        super(CompDefGoalBuilder, self).__init__(comp_child_builders.diag_store)
        self.goals: List[components.Goal] = []
        self.comp_child_builders = comp_child_builders
        self.goal_kind: Optional[str] = None

    def new_goal_header(self, goal_kind: "Token"):
        """New goal header line found.

        Arguments:
            goal_kind: Kind of goals that will follow.
        """
        if goal_kind.tok_type == "GOAL_REQUIREMENT_KW":
            self.goal_kind = components.REQUIREMENT
        else:
            assert goal_kind.tok_type == "GOAL_CONSTRAINT_KW"
            self.goal_kind = components.CONSTRAINT

    def add_goal(self, goal: components.Goal):
        """New goal has been found by the parser, add it to the found goals.

        Arguments:
            goal: Goal to add.
        """
        assert self.goal_kind is not None
        goal.goal_kind = self.goal_kind
        self.goals.append(goal)

    def add_goal_subclause(self, sub_clause: components.SubClause):
        """Subclause of the last goal has been found, add it to the last goal."""
        self.goals[-1].sub_clauses.append(sub_clause)

    def finish_comp(self, comp_def: components.ComponentDefinition, spec: "Specification"):
        """Check the found goals, and add them to the component."""
        vps = utils.construct_var_param_map(comp_def)
        cinsts = utils.construct_comp_instances_map(comp_def)
        vpps = utils.construct_verb_prepos_combis(spec)

        expr_checker = ExprChecker(vps, self.diag_store)

        # Verify all goals in the component.
        good_goals = []  # Goals without fatal error.
        elements_by_label: Dict[str, List[Any]] = self.comp_child_builders.elements_by_label
        for goal in self.goals:
            is_good = True
            assert goal.goal_kind is not None
            self.check_form("goal", goal.goal_kind, goal.doesaux, goal.sub_clauses)

            # Check existence of active and passive components.
            goal.active_comp = self.resolve_component(goal.active, cinsts)
            goal.passive_comp = self.resolve_component(goal.passive, cinsts)
            if not goal.active_comp or not goal.passive_comp:
                is_good = False

            # Store goal on its label for duplicate label detection.
            elements_by_label[goal.label_tok.tok_text].append(goal)

            if not self.verify_flows(goal.flows, vps):
                is_good = False

            self.verify_verb_prepos(goal.verb, goal.prepos, vpps)

            # Verify subclauses.
            for sub in goal.sub_clauses:
                if not expr_checker.check_expr(sub.expr):
                    is_good = False

            if is_good:
                good_goals.append(goal)

        comp_def.goals = good_goals



"""Variable groups in a component."""
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set

from raesl.compile import diagnostics
from raesl.compile.ast.components import VariableGroup
from raesl.compile.ast.nodes import GroupNode, Node
from raesl.compile.typechecking import utils
from raesl.compile.typechecking.orderer import Orderer
from raesl.utils import get_first_namepart

if TYPE_CHECKING:
    from raesl.compile.ast.components import ComponentDefinition
    from raesl.compile.ast.specification import Specification
    from raesl.compile.scanner import Token
    from raesl.compile.typechecking.compdef_builder import CompDefChildBuilders


class CollectedVarGroup:
    """Temporary data storage of found variable groups."""

    def __init__(self, name_tok: "Token", child_name_toks: List["Token"]):
        self.name_tok = name_tok
        self.child_name_toks = child_name_toks


class CompDefVarGroupBuilder:
    """Collect and check variable groups of a component definition.

    Arguments:
        comp_child_builders: Storage of child builders for a component definition.

    Attributes:
        collected_var_groups: Collected var group instances in the component.
        last_vgroup: Link to last added instance to allow adding instance arguments.
    """

    def __init__(self, comp_child_builders: "CompDefChildBuilders"):
        self.diag_store = comp_child_builders.diag_store
        self.collected_var_groups: List[CollectedVarGroup] = []
        self.comp_child_builders = comp_child_builders
        self.last_vgroup: Optional[CollectedVarGroup] = None

    def new_vargroup(self, name_tok: "Token"):
        """Parser found the start of a new variable group definition. Create a new group
        for it.

        Arguments:
            name_tok: Name of the new variable group.
        """
        self.last_vgroup = CollectedVarGroup(name_tok, [])
        self.collected_var_groups.append(self.last_vgroup)

    def vgroup_add_vars(self, varpart_name_toks: List["Token"]):
        """Parser found a line with variables that are part of the group. Store them for
        further processing afterwards.

        Arguments:
            varpart_name_toks: Name of variable parts that should become included in
                the last defined variable group.
        """
        assert self.last_vgroup is not None
        self.last_vgroup.child_name_toks.extend(varpart_name_toks)

    def finish_comp(self, comp_def: "ComponentDefinition", _spec: "Specification"):
        """Check the collected variable groups, report errors, and add the
        instances to the given component.

        Arguments:
            comp_def: Component definition to extend with the found variable groups.
                Also a source of available variables and parameters.
            spec: Specification being constructed. Source for types and relation
                definitions processed previously.
        """
        self.last_vgroup = None

        avail_varsparams = utils.construct_var_param_map(comp_def)

        # Check for duplicates and conflicts with variables and parameters.
        elements_by_label: Dict[str, List[Any]] = self.comp_child_builders.elements_by_label

        groups_by_label: Dict[str, List[CollectedVarGroup]] = defaultdict(list)

        for cgroup in self.collected_var_groups:
            elements_by_label[cgroup.name_tok.tok_text].append(cgroup)
            groups_by_label[cgroup.name_tok.tok_text].append(cgroup)
        for name, vgrps in groups_by_label.items():
            varparam = avail_varsparams.get(name)
            if varparam is not None:
                vp_loc = varparam.name_tok.get_location()
                vg_locs = [vgrp.name_tok.get_location() for vgrp in vgrps]
                kind = {True: "variable", False: "parameter"}[varparam.is_variable]
                self.diag_store.add(
                    diagnostics.E209(name, kind, "variable group", location=vp_loc, others=vg_locs)
                )

        # Build a dependency graph for the variable groups (as groups may include other
        # groups).
        orderer = Orderer()
        for cgroup in self.collected_var_groups:
            # Build a set required variables, parameters, and variable groups, using
            # non-dotted prefixes.
            needs = set(get_first_namepart(vtok.tok_text) for vtok in cgroup.child_name_toks)
            orderer.add_dependency(cgroup.name_tok.tok_text, needs, cgroup)

        reported: Set[str] = set()  # Reported failures
        vargroups: Dict[str, VariableGroup] = {}  # Resolved variable groups.

        resolved, cycle = orderer.resolve()
        for entry in resolved:
            cgroup = entry.data
            if cgroup is None:
                # Need entry added by the orderer, ignore.
                continue

            # Next collected group to deal with.
            #
            # Each of its variablepart_names either contains non-dotted previous
            # variable groups, or possibly dotted variables or parameters.
            content_vgroup: List[Node] = []  # Result content for the future variable group.
            for partname_tok in cgroup.child_name_toks:
                node = utils.resolve_var_param_group_node(
                    partname_tok,
                    avail_varsparams,
                    vargroups,
                    reported,
                    self.diag_store,
                )
                if node is not None:
                    content_vgroup.append(node)
                # Else, error already reported, ignore it.

            # Even with errors a variable group is constructed to avoid follow-up
            # errors on non-existing groups.
            vgroup = VariableGroup(cgroup.name_tok, cgroup.child_name_toks)
            vgroup.node = GroupNode(cgroup.name_tok, content_vgroup)
            vargroups[cgroup.name_tok.tok_text] = vgroup
            comp_def.var_groups.append(vgroup)

        if cycle:
            locs = [entry.data.name_tok.get_location() for entry in cycle]
            name = cycle[0].data.name_tok.tok_text
            self.diag_store.add(
                diagnostics.E204(name, "variable group", location=locs[0], cycle=locs)
            )



"""Relation definition and instantiation."""
from typing import TYPE_CHECKING, List, Optional

from raesl.compile.ast.comment_storage import DefaultDocStore

if TYPE_CHECKING:
    from raesl.compile.ast.types import BaseType
    from raesl.compile.scanner import Token

INPUT = "input"
OUTPUT = "output"
INPOUT = "inp_out"


class RelationDefParameter:
    """Parameter of a relation definition.

    Arguments:
        name: Name of the parameter.
        type_name: Name of the type of the parameter.
        direction: Direction of the parameter.
        multi: If set, parameter may be specified more than once.

    Attributes:
        type: Actual type of the parameter.
    """

    def __init__(self, name: "Token", type_name: "Token", direction: str, multi: bool):
        self.name = name
        self.type_name = type_name
        self.direction = direction
        self.multi = multi

        self.type: Optional["BaseType"] = None


class RelationDefinition(DefaultDocStore):
    """Relation definition.

    Arguments:
        name: Name of the relation definition.

    Attributes:
        params: Parameters of the definition.
    """

    def __init__(self, name: "Token"):
        super(RelationDefinition, self).__init__(name)
        self.name = name
        self.params: List[RelationDefParameter] = []



"""Default options for ESL to Excel workbook conversion."""
from openpyxl.styles import Alignment, Font
from openpyxl.worksheet.protection import SheetProtection

WRAP = Alignment(wrap_text=True)
MONO = Font(name="Consolas")
SHEETPROTECTION = SheetProtection(
    sheet=True,
    formatColumns=False,
    formatRows=False,
    formatCells=False,
    sort=False,
    autoFilter=False,
)

OPTIONS = {
    # General
    "instance path": dict(
        styles=dict(alignment=Alignment(horizontal="right", vertical="top", wrap_text=True)),
        width=30,
    ),
    "component path": dict(
        styles=dict(alignment=Alignment(horizontal="right", vertical="top", wrap_text=True)),
        width=40,
    ),
    "instance name": dict(
        styles=dict(alignment=Alignment(horizontal="right", vertical="top", wrap_text=True)),
        width=20,
    ),
    "parent component": dict(
        styles=dict(alignment=Alignment(horizontal="right", vertical="top", wrap_text=False)),
        width=30,
    ),
    "component definition": dict(
        styles=dict(alignment=Alignment(horizontal="left", vertical="top", wrap_text=True)),
        width=30,
    ),
    "kind": dict(styles=dict(alignment=Alignment(horizontal="center", vertical="top")), width=20),
    "form": dict(styles=dict(alignment=Alignment(horizontal="center", vertical="top")), width=15),
    # Helper words
    "auxiliary": dict(
        styles=dict(alignment=Alignment(horizontal="center", vertical="top", wrap_text=True)),
        width=12,
    ),
    "comparison": dict(
        styles=dict(alignment=Alignment(horizontal="center", vertical="top", wrap_text=True)),
        width=20,
    ),
    "comments": dict(
        styles=dict(alignment=Alignment(horizontal="left", vertical="top", wrap_text=True)),
        width=60,
    ),
    # Overview
    "specification text": dict(
        styles=dict(alignment=Alignment(horizontal="left", vertical="top", wrap_text=True)),
        width=60,
    ),
    # Need
    "subject": dict(
        styles=dict(alignment=Alignment(horizontal="left", vertical="top", wrap_text=True)),
        width=30,
    ),
    # Goal/transform specific
    "source component": dict(
        styles=dict(alignment=Alignment(horizontal="right", vertical="top", wrap_text=True)),
        width=40,
    ),
    "target component": dict(
        styles=dict(alignment=Alignment(horizontal="right", vertical="top", wrap_text=True)),
        width=40,
    ),
    "verb": dict(
        styles=dict(alignment=Alignment(horizontal="center", vertical="top", wrap_text=True)),
        width=15,
    ),
    "preposition": dict(
        styles=dict(alignment=Alignment(horizontal="center", vertical="top", wrap_text=True)),
        width=10,
    ),
    "variables": dict(
        styles=dict(alignment=Alignment(horizontal="left", vertical="top", wrap_text=True)),
        width=30,
    ),
    # Design requirements
    "bound": dict(
        styles=dict(alignment=Alignment(horizontal="right", vertical="top", wrap_text=True)),
        width=20,
    ),
    "subclauses": dict(
        styles=dict(alignment=Alignment(horizontal="left", vertical="top")), width=60
    ),
}

OUTPUT = "./esl.xlsx"
SCOPES = dict(world=None)



"""Line matching state machines for groups."""
from typing import TYPE_CHECKING

from raesl.compile.machine_files import argument_list, typing

if TYPE_CHECKING:
    from raesl.compile.typechecking.ast_builder import AstBuilder


_GROUP_SECTION_HEADER_SPEC = """
group_section_header:
    start initial;
    start -> s1 [VARIABLE_GROUP_KW] tag=vargroup;

    end accept=group_section_header;
    s1 -> end [NL_TK];
    s1 -> end [EOF_TK];
"""


def _process_vargroup_header(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    builder.notify_new_section(tags["vargroup"][0], False)


_GROUP_START_SPEC = """
group_start:
    start initial;
    start -> s1 [NAME] tag=group_name;
    s1 -> s2 [IS_KW];
    s2 -> s3 [A_KW];
    s3 -> s4 [GROUP_KW];
    s4 -> s5 [OF_KW];

    end accept=group_start;
    s5 -> end [NL_TK];
    s5 -> end [EOF_TK];
"""


def _process_new_vargroup(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.new_vargroup(tags["group_name"][0])


def _process_arguments(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    names = argument_list.process_argument_list_line(tags)
    current_comp.vgroup_add_vars(names)


MACHINES: typing.MachineTripletList = [
    (
        "GROUP_SECTION_HEADER_MACHINE",
        _GROUP_SECTION_HEADER_SPEC,
        _process_vargroup_header,
    ),
    ("GROUP_START_MACHINE", _GROUP_START_SPEC, _process_new_vargroup),
    (
        "GROUP_ARGUMENT_LINE_MACHINE",
        argument_list.ARGUMENT_LINE_SPEC,
        _process_arguments,
    ),
]



"""Code for collecting and type checking designs."""
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from raesl.compile.ast import components
from raesl.compile.typechecking import utils
from raesl.compile.typechecking.expr_checker import ExprChecker

if TYPE_CHECKING:
    from raesl.compile.ast.specification import Specification
    from raesl.compile.scanner import Token
    from raesl.compile.typechecking.compdef_builder import CompDefChildBuilders


class CompDefDesignBuilder:
    """Class for collecting and type checking designs in a component definition."""

    def __init__(self, comp_child_builders: "CompDefChildBuilders"):
        self.diag_store = comp_child_builders.diag_store
        self.comp_child_builders = comp_child_builders
        self.designs: List[components.Design] = []
        self.design_kind: Optional[str] = None

    def new_design_header(self, kind: "Token"):
        """New design section started, store the kind stated in the header."""
        if kind.tok_type == "DESIGN_REQUIREMENT_KW":
            self.design_kind = components.REQUIREMENT
        else:
            assert kind.tok_type == "DESIGN_CONSTRAINT_KW"
            self.design_kind = components.CONSTRAINT

    def design_line(self, design: components.Design):
        """New design rule found, store it."""
        assert self.design_kind is not None
        design.design_kind = self.design_kind
        self.designs.append(design)

    def add_design_subclause(self, sub: components.SubClause):
        """Subclause of the last design has been found, append it to the last design."""
        self.designs[-1].sub_clauses.append(sub)

    def finish_comp(self, comp_def: components.ComponentDefinition, _spec: "Specification"):
        """Check the found designs in the context of 'comp_def', and add them after
        verification.

        Arguments:
            comp_def: Surrounding component definition supplying variables and
                parameters. Checked designs should be added to it after checking.
            _spec: Specification being constructed, source for types and verbs.
        """
        vps = utils.construct_var_param_map(comp_def)
        expr_checker = ExprChecker(vps, self.diag_store)

        good_designs = []  # Designs that can be added.
        elements_by_label: Dict[str, List[Any]] = self.comp_child_builders.elements_by_label
        for design in self.designs:
            is_good = True

            # Order by label for double use checking.
            elements_by_label[design.label_tok.tok_text].append(design)

            # Verify comparisons.
            if not expr_checker.check_expr(design.expr):
                is_good = False

            for sub in design.sub_clauses:
                if not expr_checker.check_expr(sub.expr):
                    is_good = False

            # Consider design good if no big issues found.
            if is_good:
                good_designs.append(design)

        comp_def.designs = good_designs



from copy import deepcopy
from typing import Generator, List, Optional

import raesl.doc.lines as lns
from raesl.doc import utils
from raesl.doc.locales import _, pm
from ragraph.graph import Graph
from ragraph.node import Node

LineGen = Generator[str, None, None]
TABLE_SINGLE = (
    "+------------------------------------------------"
    "------------------------------------------------+"
)
TABLE_LEFT_DOUBLE = (
    "+:==============================================="
    "================================================+"
)


def node_decomp_level(
    depth: int, comps: List[Node], h: int = 1, rich: str = None, rich_opts: dict = {}
):
    """Yield a decomposition level intro section."""
    yield lns.header(
        h,
        "{} {} \\label{{{}}}".format(
            _("system specification decomposition level"),
            depth + 1,
            "chp:esl" + str(depth + 1),
        ),
    )
    if depth == 0:
        leader = lns.snt(
            _(
                "this chapters describes the system of interest at the first "
                "decomposition level. That is, it describes {} components which play a "
                "role within the environment in which the system of interest must "
                "operate and the (functional) interactions between those components"
            ).format(len(comps))
        )
    elif depth > 0 and len(comps) == 1:
        leader = lns.snt(
            _(
                "this chapters describes the system of interest at decomposition level "
                "{} and introduces one additional component"
            ).format(depth + 1)
        )
    else:
        leader = lns.snt(
            _(
                "this chapters describes the system of interest at decomposition level "
                "{} and introduces {} additional components"
            ).format(depth + 1, len(comps))
        )
    yield leader

    if rich == "tex" or rich == "md":
        yield lns.snt(
            _(
                "in Figure \\ref{{{}}} the associated design-structure-matrix (DSM) is " "shown"
            ).format("fig:mdmlevel" + str(depth + 1))
        )
        yield lns.snt(
            _(
                "the DSM shows the dependencies between the elements that are relevant "
                "to this decomposition level"
            )
        )


def global_needs_and_designs(h: int = 1):
    """Yield a global need and design specification intro section."""
    yield lns.header(h, "{}".format(_("system-wide qualitative and quantitative specifications")))

    yield lns.snt(
        _(
            "this chapter lists all qualitative and quantitative design specifications "
            "that cannot be linked to a component"
        )
    )


def get_node_table(
    node: Node, graph: Optional[Graph] = None, sub_nodes: Optional[List[str]] = None
) -> List:
    """Returns a Markdown grid table."""
    table = []
    table.append(TABLE_SINGLE)
    table.append("| {} |".format(lns.bold(lns.node_path(node.name))))
    table.append(TABLE_LEFT_DOUBLE)
    lines = list(lns.lines(node, graph=graph))
    main_clause = lines[0].replace("\n", "")
    table.append("| {} |".format(main_clause))
    if len(lines) > 1:
        for line in lines[1:]:
            table.append("| {} |".format(line.replace("\n", "")))
        table.append("|    |")
    table.append(TABLE_SINGLE)

    if sub_nodes:
        table.append(
            "| {} |".format(
                lns.boldhead(_("subordinate function specifications")).replace("\n", "")
            )
        )
        table.append(TABLE_SINGLE)
        for s in sub_nodes:
            table.append("| {} |".format(lns.node_path(s).replace("\n", "")))
        table.append("|    |")

        table.append(TABLE_SINGLE)

    plain_comments = (
        [("comments", node.annotations.esl_info.get("comments", []))]
        if node.annotations.esl_info.get("comments", [])
        else []
    )

    for key, comments in plain_comments + list(
        node.annotations.esl_info["tagged_comments"].items()
    ):
        table.append("| {} |".format(lns.boldhead(_(key)).replace("\n", "")))
        table.append(TABLE_SINGLE)
        for comment in comments:
            table.append("| {} |".format(comment.replace("\n", "")))
        table.append(TABLE_SINGLE)

    table.append("")

    return table


# Goal specs
def comp_node_goal_reqs(node: Node, g: Graph, h: int = 1) -> LineGen:
    """Yield a goal-requirement section for a component"""
    glsn = utils.get_component_goals(node, g, constraint=False, inherited=False)
    glsa = utils.get_component_goals(node, g, constraint=False, inherited=True)

    if not glsn and not glsa:
        return

    yield lns.header(h, _("goal function requirements"))

    for gl in glsn:
        table = get_node_table(node=gl, graph=g)
        yield from table

    for gl in glsa:
        temp = deepcopy(gl)
        temp.annotations.esl_info["comments"].append(
            lns.snt(
                _("this goal function requirement automatically migrated from {}").format(
                    lns.node_path(temp.annotations.esl_info["body"]["active"])
                )
            )
        )
        temp.annotations.esl_info["body"]["active"] = node.name
        table = get_node_table(node=temp, graph=g)
        yield from table


def comp_node_goal_cons(node: Node, g: Graph, h: int = 1) -> LineGen:
    """Yield a goal-constraint section for a component"""
    glsn = utils.get_component_goals(node, g, constraint=True, inherited=False)
    glsa = utils.get_component_goals(node, g, constraint=True, inherited=True)
    if not glsn and not glsa:
        return

    yield lns.header(h, _("goal function constraints"))

    for gl in glsn:
        table = get_node_table(node=gl, graph=g)
        yield from table

    for gl in glsa:
        temp = deepcopy(gl)
        temp.annotations.esl_info["comments"].append(
            lns.snt(
                _("this goal function constraint automatically migrated from {}").format(
                    lns.node_path(temp.annotations.esl_info["body"]["active"])
                )
            )
        )
        temp.annotations.esl_info["body"]["active"] = node.name
        table = get_node_table(node=temp, graph=g)
        yield from table


# Transformation specs
def comp_node_transformation_reqs(node: Node, g: Graph, h: int = 1) -> LineGen:
    """Yield a transformation requirements section for a component."""
    trs = utils.get_component_transformations(node, g, constraint=False)
    if not trs:
        return

    trs = sorted(trs, key=lambda x: lns.node_path(x.name))

    yield lns.header(h, _("transformation function requirements"))

    for tr in trs:
        subs = set(
            [
                e.target.name
                for e in g.edges
                if e.source == tr and e.kind == "traceability_dependency"
            ]
        )
        table = get_node_table(tr, graph=g, sub_nodes=subs)
        yield from table


def comp_node_transformation_cons(node: Node, g: Graph, h: int = 1) -> LineGen:
    """Yield a transformation constraints section for a component."""
    trs = utils.get_component_transformations(node, g, constraint=True)
    if not trs:
        return

    trs = sorted(trs, key=lambda x: lns.node_path(x.name))

    yield lns.header(h, _("transformation function constraints"))

    for tr in trs:
        subs = [
            e.target.name for e in g.edges if e.source == tr and e.kind == "traceability_dependency"
        ]
        table = get_node_table(tr, graph=g, sub_nodes=subs)
        yield from table


# Behavior specs
def comp_node_behavior_reqs(node: Node, g: Graph, h: int = 1) -> LineGen:
    """Yield a behavior requirement section for a component node."""
    bhs = utils.get_component_behaviors(node, g, constraint=False)
    if not bhs:
        return

    yield lns.header(h, _("behavior requirements"))

    for b in bhs:
        table = get_node_table(node=b, graph=g)
        yield from table


# Behavior specs
def comp_node_behavior_cons(node: Node, g: Graph, h: int = 1) -> LineGen:
    """Yield a behavior constraint section for a component node."""
    bhs = utils.get_component_behaviors(node, g, constraint=True)
    if not bhs:
        return

    yield lns.header(h, _("behavior constraints"))

    for b in bhs:
        table = get_node_table(node=b, graph=g)
        yield from table


# Design specs
def comp_node_design_reqs(node: Node, g: Graph, h: int = 1) -> LineGen:
    """Yield a design requirements section for a component node."""
    drs = utils.get_component_designs(node, g, constraint=False)
    if not drs:
        return

    yield lns.header(h, _("quantitative design requirements"))

    for d in drs:
        table = get_node_table(node=d, graph=g)
        yield from table


# Design specs
def global_design_reqs(g: Graph, h: int = 1) -> LineGen:
    """Yield a global design requirements section."""
    drs = utils.get_global_designs(g, constraint=False)
    if not drs:
        return

    yield lns.header(h, _("quantitative design requirements"))

    for d in drs:
        table = get_node_table(node=d, graph=g)
        yield from table


def comp_node_design_cons(node: Node, g: Graph, h: int = 1) -> LineGen:
    """Yield a design constraint section for a component node."""
    dcs = utils.get_component_designs(node, g, constraint=True)
    if not dcs:
        return

    yield lns.header(h, _("quantitative design constraints"))

    for d in dcs:
        table = get_node_table(node=d, graph=g)
        yield from table


# Design specs
def global_design_cons(g: Graph, h: int = 1) -> LineGen:
    """Yield a global design constraint section."""
    dcs = utils.get_global_designs(g, constraint=True)
    if not dcs:
        return

    yield lns.header(h, _("quantitative design constraints"))

    for d in dcs:
        table = get_node_table(node=d, graph=g)
        yield from table


# Needs
def comp_node_needs(node: Node, g: Graph, h: int = 1) -> LineGen:
    """Yield a need section for a component node."""
    nds = utils.get_component_needs(node, g)
    if not nds:
        return

    yield lns.header(h, _("qualitative design requirements"))

    for n in nds:
        table = get_node_table(node=n, graph=g)
        yield from table


def global_needs(g: Graph, h: int = 1) -> LineGen:
    """Yield a global need section."""
    nds = utils.get_global_needs(g)
    if not nds:
        return

    yield lns.header(h, _("qualitative design requirements"))

    for n in nds:
        table = get_node_table(node=n, graph=g)
        yield from table


# Relations
def relation_node_table(r: Node, g: Graph) -> List:
    ri = r.annotations.esl_info
    table = []
    table.append(TABLE_SINGLE)
    table.append("| {} |".format(lns.bold(lns.node_path(r.name))))
    table.append(TABLE_LEFT_DOUBLE)
    table.append("| {} |".format(lns.bold(_("model definition name"))))
    table.append(TABLE_SINGLE)
    table.append("|    |")
    table.append("| {} |".format(ri["definition_name"].replace("\n", "")))
    table.append("|    |")
    table.append(TABLE_SINGLE)

    if ri.get("required_variables"):
        table.append("| {} |".format(lns.bold(_("required variables"))))
        table.append(TABLE_SINGLE)
        for v in ri.get("required_variables"):
            table.append("|    |")
            table.append("| {} |".format(lns.var_path(g[v]).replace("\n", "")))
            table.append("|    |")
        table.append(TABLE_SINGLE)

    if ri.get("returned_variables"):
        table.append("| {} |".format(lns.bold(_("returned variables"))))
        table.append(TABLE_SINGLE)
        for v in ri.get("returned_variables"):
            table.append("|    |")
            table.append("| {} |".format(lns.var_path(g[v]).replace("\n", "")))
            table.append("|    |")
        table.append(TABLE_SINGLE)

    if ri.get("related_variables"):
        table.append("| {} |".format(lns.bold(_("related variables"))))
        table.append(TABLE_SINGLE)
        for v in ri.get("related_variables"):
            table.append("|    |")
            table.append("| {} |".format(lns.var_path(g[v]).replace("\n", "")))
            table.append("|    |")
        table.append(TABLE_SINGLE)

    plain_comments = [("comments", ri.get("comments", []))] if ri.get("comments", []) else []

    for key, comments in plain_comments + list(ri["tagged_comments"].items()):
        table.append("| {} |".format(lns.boldhead(_(key)).replace("\n", "")))
        table.append(TABLE_SINGLE)
        for comment in comments:
            table.append("| {} |".format(comment.replace("\n", "")))
        table.append(TABLE_SINGLE)

    table.append("")
    return table


def comp_node_relations(node: Node, g: Graph, h: int = 1) -> LineGen:
    """Yield a relation section for a component."""
    rls = [n for n in g.nodes if n.kind == "relation_spec"]

    c_rls = []
    for r in rls:
        if g[node.name, r.name]:
            c_rls.append(r)

    if not c_rls:
        return

    c_rls = sorted(c_rls, key=lambda x: lns.node_path(x.name))

    yield lns.header(h, _("external models"))

    for r in c_rls:
        table = relation_node_table(r=r, g=g)
        yield from table


def comp_node_props(node: Node, g: Graph, h: int) -> LineGen:
    """Yield a properties section for a component node"""
    props = utils.get_component_properties(node, g)
    if not props:
        return

    name = node.name.split(".")[-1]
    yield lns.cap(_("the following properties are specified for {}").format(name)) + ":"
    yield from lns.unordered(sorted([p.name.split(".")[-1] for p in props]))


def comp_node_subcomps(node: Node, h: int = 1):
    """Yield a subcomponents sections for a component."""
    if not node.children:
        return

    yield lns.header(h, _("sub-components"))

    child_names = [c.name.split(".")[-1] for c in node.children]

    name = node.name.split(".")[-1]

    yield lns.cap(_("{} is composed of the following sub-components:").format(name))

    yield from lns.unordered(sorted(child_names))


def var_node_table(g: Graph, h: int = 1, reference_list: bool = False) -> LineGen:
    """Yield a variable table section."""
    vrs = [n for n in g.nodes if n.kind == "variable"]
    if not vrs:
        return

    vrs = sorted(vrs, key=lambda x: lns.node_path(x.name))

    yield lns.header(h, _("list of variables"))

    yield lns.header(h + 1, _("definitions"))

    # yield "\\begin{landscape}"
    yield ""
    yield "| {} | {} | {} | {} | {} |".format(
        lns.boldhead(_("variable"), newlines=False),
        lns.boldhead(_("type"), newlines=False),
        lns.boldhead(_("domain"), newlines=False),
        lns.boldhead(_("units"), newlines=False),
        lns.boldhead(_("clarification"), newlines=False),
    )
    yield "|:---|:---|:---|:---|:---|"

    for v in vrs:
        vname, tname, domain, units, comments = get_var_table_row_elements(g, v)
        yield "| {} | {} | {} | {} | {} |".format(
            lns.var_path(v), lns.cap(tname), domain, units, comments
        )
    yield ""
    # yield "\\end{landscape}"

    if reference_list:
        yield "\\newpage{}"
        yield lns.header(h + 1, _("variable reference list"))

        yield "| {} | {} |".format(
            lns.boldhead(_("variable"), newlines=False),
            lns.boldhead(_("related to"), newlines=False),
        )
        yield "|:---|:---|"
        for v in vrs:
            rels = [
                lns.node_path(e.source.name)
                for e in g.edges
                if e.kind == "mapping_dependency" and e.target == v
            ]
            rels += [
                lns.node_path(e.target.name)
                for e in g.edges
                if e.kind == "mapping_dependency" and e.source == v
            ]
            if rels:
                refs = pm.hook.linguistic_enumeration(items=rels)
            else:
                refs = ""
            yield "| {} | {} |".format(lns.node_path(v.name), refs)
        yield ""


def get_var_table_row_elements(g: Graph, v: Node) -> str:
    """Yield elements of a row in a var table."""
    if v.annotations.esl_info["comments"]:
        comments = " ".join(v.annotations.esl_info["comments"])
    else:
        comments = ""

    for key, value in v.annotations.esl_info["tagged_comments"].items():
        comments += " ".join([" ", lns.bold(lns.cap(key + ":"))] + value)

    tname = v.annotations.esl_info["type_ref"]
    t = g[tname]
    if t.annotations.esl_info.get("domain"):
        intervals = []
        for ival in t.annotations.esl_info["domain"]:
            if ival["lowerbound"]["value"] != ival["upperbound"]["value"]:
                is_enum = False

                iv = ""
                if ival["lowerbound"]["value"]:
                    unit = ival["lowerbound"]["unit"]
                    if not unit:
                        unit = ""
                    iv += "{} {} $\\leq$ x".format(ival["lowerbound"]["value"], unit)

                if ival["upperbound"]["value"] and ival["lowerbound"]["value"]:
                    unit = ival["upperbound"]["unit"]
                    if not unit:
                        unit = ""
                    iv += "$\\leq$ {} {}".format(ival["upperbound"]["value"], unit)
                elif ival["upperbound"]["value"] and not ival["lowerbound"]["value"]:
                    unit = ival["upperbound"]["unit"]
                    if not unit:
                        unit = ""
                    iv += "x $\\leq$ {} {}".format(ival["upperbound"]["value"], unit)
                intervals.append(iv)
            else:
                is_enum = True
                value = ival["lowerbound"]["value"]
                if ival["lowerbound"]["unit"]:
                    value += " " + ival["lowerbound"]["unit"]
                intervals.append(value)
        if not is_enum:
            domain = ", ".join(intervals)
        else:
            domain = _("enumeration of ") + pm.hook.linguistic_enumeration(items=intervals)
    else:
        domain = ""
    if t.annotations.esl_info.get("units"):
        units = ",".join(t.annotations.esl_info.get("units"))
    else:
        units = ""

    return v.name, tname, domain, units, comments



"""Code for handling 'needs' in component definitions."""
from typing import TYPE_CHECKING, Any, Dict, List

from raesl.compile import diagnostics
from raesl.compile.ast.components import Need
from raesl.compile.ast.nodes import ElementaryVarNode
from raesl.compile.typechecking import utils
from raesl.compile.typechecking.utils import resolve_var_param_node

if TYPE_CHECKING:
    from raesl.compile.ast.components import ComponentDefinition
    from raesl.compile.ast.specification import Specification
    from raesl.compile.scanner import Token
    from raesl.compile.typechecking.compdef_builder import CompDefChildBuilders


class CompDefNeedBuilder:
    """Class for handling 'need' sections.

    Arguments:
        comp_child_builders: Storage of child builders for a component definition.

    Attributes:
        needs: Collected needs.
    """

    def __init__(self, comp_child_builders: "CompDefChildBuilders"):
        self.diag_store = comp_child_builders.diag_store
        self.comp_child_builders = comp_child_builders

        self.needs: List[Need] = []

    def add_need(self, label_tok: "Token", subject_tok: "Token", description: str):
        """Parser found another need, store it for future processing."""
        need = Need(label_tok, subject_tok, description)
        self.needs.append(need)

    def finish_comp(self, comp_def: "ComponentDefinition", _spec: "Specification"):
        """Check the collected needs, and store them in the component definition."""
        # Verify non-duplicated need labels.
        elements_by_label: Dict[str, List[Any]] = self.comp_child_builders.elements_by_label
        for need in self.needs:
            elements_by_label[need.label_tok.tok_text].append(need)

        # Construct link to the definition of the subject mentioned in the neEd.
        # If found, add the need to the component definition, else give an error.
        vps = utils.construct_var_param_map(comp_def)
        cinsts = utils.construct_comp_instances_map(comp_def)
        rgtdbs = utils.construct_relinst_goal_transform_design_behavior_map(comp_def)

        for need in self.needs:
            cinst = cinsts.get(need.subject_tok.tok_text)
            if cinst is not None:
                need.subject = cinst
                comp_def.needs.append(need)
                continue

            rgtdb = rgtdbs.get(need.subject_tok.tok_text)
            if rgtdb is not None:
                # XXX Returned type of rgtdb looks like a subset of allowed subject
                # types. Check!
                need.subject = rgtdb
                comp_def.needs.append(need)
                continue

            varparam = resolve_var_param_node(need.subject_tok, vps, set(), self.diag_store)
            if varparam is not None:
                if isinstance(varparam, ElementaryVarNode):
                    need.subject = varparam
                    comp_def.needs.append(need)
                    continue
                else:
                    loc = need.subject_tok.get_location()
                    name = need.label_tok.tok_text
                    self.diag_store.add(diagnostics.E226(name, location=loc))
                    continue

            loc = need.subject_tok.get_location()
            name = need.subject_tok.tok_text
            context = need.label_tok.tok_text
            self.diag_store.add(
                diagnostics.E205(
                    f"subject '{name}'",
                    f"the context of need '{context}'",
                    location=loc,
                )
            )



"""Rich document content."""
from pathlib import Path
from typing import Any, Dict, Generator, Optional, Union

import raesl.plot
from raesl import logger
from raesl.doc.locales import _
from ragraph.graph import Graph

IMG_DIR = Path("./images")

DEFAULT_NODE_KINDS = ["component", "function_spec"]
DEFAULT_EDGE_KINDS = ["functional_dependency", "mapping_dependency"]
DEFAULT_PIE_MODE = "relative"


def mdm(
    graph: Graph,
    depth: int,
    rich: str = "tex",
    rich_opts: Dict[str, Any] = {},
    img_dir: Optional[Union[Path, str]] = None,
) -> Generator[str, None, None]:
    """Generate an Multi-Domain Matrix.


    Yields:
        Rich output lines.
    """
    logger.debug("Generating MDM for depth {}...".format(depth))

    img_dir = Path(img_dir if img_dir is not None else rich_opts.get("img_dir", IMG_DIR))
    img_dir.mkdir(parents=True, exist_ok=True)

    level = depth + 1

    style = raesl.plot.Style(
        ragraph=dict(
            piemap={
                "display": "labels",
                "mode": rich_opts.get("pie_mode", DEFAULT_PIE_MODE),
            }
        )
    )

    fig = raesl.plot.mdm(
        graph,
        node_kinds=rich_opts.get("node_kinds", DEFAULT_NODE_KINDS),
        edge_kinds=rich_opts.get("edge_kinds", DEFAULT_EDGE_KINDS),
        depth=level,
        style=style,
    )

    # Width of a pixel in mm on A4 paper with 96 dpi print quality.
    pix_in_mm = 0.26458
    # Width of a line on A4 paper with 20 mm side margins
    line_width = 170

    if fig.layout.width * pix_in_mm < line_width:
        fig_size = "width={:.3f}\\linewidth".format(fig.layout.width * pix_in_mm / line_width)
        angle = 0
    else:
        fig_size = "height={:.3f}\\linewidth".format(
            min(1.0, fig.layout.height * pix_in_mm / line_width)
        )
        angle = 90

    caption = _("{} dependency matrix of decomposition level {}.").format(
        " -- ".join(rich_opts.get("node_kinds", DEFAULT_NODE_KINDS)), level
    )

    img_path = str(img_dir / "level-{}".format(level))
    if not img_dir.is_absolute():
        img_path == "./" + img_path

    logger.debug("Generating image using Plotly's Kaleido...")
    try:
        import platform

        machine = platform.machine()
        if "arm" in machine or "aarch" in machine:
            import plotly.io as pio

            if "--single-process" not in pio.kaleido.scope.chromium_args:
                pio.kaleido.scope.chromium_args += ("--single-process",)

        if rich == "tex":
            img_path = f"{img_path}.pdf"
            fig.write_image(img_path)
            latex_path = img_path.replace("\\", "/").replace("_", "-underscore-")
            yield "\\begin{figure}[!htbp]"
            yield "\\centering"
            yield "\\includegraphics[{}, angle={}]{{{}}}".format(fig_size, angle, latex_path)
            yield "\\caption{{{}}}\\label{{{}}}".format(caption, "fig:mdmlevel" + str(level))
            yield "\\end{figure}"
            logger.debug("Included image as {}.".format(rich))

        elif rich == "md":
            img_path = f"{img_path}.svg"
            fig.write_image(img_path)
            yield "![{}\\label{{{}}}]({})".format(
                caption,
                "fig:mdmlevel" + str(level),
                img_path.replace("\\", "/").replace("_", "-underscore-"),
            )
            yield "\n"
            logger.debug("Included image as {}.".format(rich))
    except Exception as e:
        logger.error(
            "Something went wrong when generating images using Plotly's Kaleido. "
            + "Kaleido might not be available. "
            + "You can also disable rich output generation for now. "
        )
        raise e



"""Generic class for resolving an order of checking 'things' assuming each thing
has a name, and knows the names of its direct dependencies.
"""
from typing import Dict, Iterable, List, Optional, Tuple

TypeOfData = Optional[None]


class OrdererEntry:
    """Entry in the orderer.

    Arguments:
        name: Name of the entry.
        data: Data associated with the entry.

    Attributes:
        needed_by: List of entries that depend on this entry. May only be accessed by
            the Orderer.
        depend_on: List of entries that need this entry. May only be accessed by
            the Orderer.
    """

    def __init__(self, name: str, data: TypeOfData) -> None:
        self.name = name
        self.data = data

        # Internal data.
        self.needed_by: List[OrdererEntry] = []
        self.depend_on: List[OrdererEntry] = []

    def __repr__(self):
        return "OrdererEntry[{}]".format(self.name)


class Orderer:
    """Class that orders entries so their 'needed_by' entries are processed first.

    Attributes:
        _independent_entries: Entries that do not depend on other entries.
        _dependent_entries: Entries that depend on other entries.
    """

    def __init__(self) -> None:
        self._independent_entries: Dict[str, OrdererEntry] = {}
        self._dependent_entries: Dict[str, OrdererEntry] = {}

    def add_dependency(self, provide: str, needs: Iterable[str], data: TypeOfData = None):
        """Add a dependency where the 'provide' name depends on the 'needs' names.

        Arguments:
            provide: Name of the 'thing' that this dependency provides.
            needs: Names of 'things' that are required by the provide 'thing'.
            data: Optional data associated with 'provide'. At most one such data item
                should exist, class cannot handle multiple data items.
        """
        provide_entry = self._find_create_entry(provide, data)
        provide_independent = len(provide_entry.depend_on) == 0

        # Add needs as 'things' it depends on.
        for need in needs:
            need_entry = self._find_create_entry(need)
            need_entry.needed_by.append(provide_entry)
            provide_entry.depend_on.append(need_entry)

        # If 'provide_entry' became dependent, move it.
        if provide_independent and len(provide_entry.depend_on) > 0:
            del self._independent_entries[provide]
            self._dependent_entries[provide] = provide_entry

    def resolve(self) -> Tuple[List[OrdererEntry], Optional[List[OrdererEntry]]]:
        """Resolve the dependency chain by peeling away independent entries. This should
        cause some dependent entries to become independent, allowing them to be peeled
        as well.

        Returns:
            Ordered sequence of entries containing name and associated data, and
                optionally a cycle of entries if at least one cycle exists. Note that
                the Orderer returns one arbitrary cycle in such a case.
        """
        ordered = []
        while len(self._independent_entries) > 0:
            entry = self._independent_entries.popitem()[1]
            assert len(entry.depend_on) == 0
            ordered.append(entry)

            # Peeling done, update dependent entries.
            while len(entry.needed_by) > 0:
                dependent = entry.needed_by.pop()
                dependent.depend_on.remove(entry)
                if len(dependent.depend_on) == 0:
                    # Dependent entry became independent!
                    del self._dependent_entries[dependent.name]
                    self._independent_entries[dependent.name] = dependent

        if len(self._dependent_entries) == 0:
            return ordered, None

        # We have at least one cycle, find one.
        def find_cycle(
            stack: List[OrdererEntry], entry_indices: Dict[OrdererEntry, int]
        ) -> Optional[List[OrdererEntry]]:
            entry = stack[-1]
            for dep in entry.needed_by:
                if dep in entry_indices:
                    # Bingo!
                    return stack[entry_indices[dep] :]
                entry_indices[dep] = len(stack)
                stack.append(dep)
                cycle = find_cycle(stack, entry_indices)
                if cycle is not None:
                    return cycle
                stack.pop()
                del entry_indices[dep]
            return None

        entry = next(iter(self._dependent_entries.values()))
        stack = [entry]
        entry_indices = {entry: 0}
        cycle = find_cycle(stack, entry_indices)
        assert cycle is not None
        return ordered, cycle

    def find_entry(self, name: str) -> Optional[OrdererEntry]:
        """Try to find an entry with the provided name. Useful for duplicate name
        detection.

        Returns:
            The found entry (treat as read-only), or None.
        """
        entry = self._independent_entries.get(name)
        if entry is not None:
            entry = self._dependent_entries.get(name)
        return entry

    def _find_create_entry(self, name: str, data: TypeOfData = None) -> OrdererEntry:
        """Find or create an entry with the provide name.

        Arguments:
            name: name of the entry to find or create.
            data: Optional data associated with the name. It is copied into the entry
                whenever possible without overwriting existing data.

        Returns:
            The entry.
        """
        entry = self._independent_entries.get(name)
        if entry is not None:
            if entry.data is None and data is not None:
                entry.data = data
            return entry

        entry = self._dependent_entries.get(name)
        if entry is not None:
            if entry.data is None and data is not None:
                entry.data = data
            return entry

        entry = OrdererEntry(name, data)
        self._independent_entries[name] = entry
        return entry



"""RaESL GraphViz diagram plotting module."""

from typing import Any, Dict, Optional, Union

from ragraph.plot.generic import Style as RaGraphStyle
from ragraph.generic import Mapping, field


class DiagramStyle(Mapping):
    """RaESL Graphviz diagram style mapping."""

    _defaults = dict(
        digraph=dict(filename="raesl.dot"),
        orientation="TD",
        node_shapes=dict(
            component="rectangle",
            goal="hexagon",
            transformation="ellipse",
        ),
        edge_styles=dict(
            functional_dependency="solid",
            logical_dependency="dashed",
            traceability_dependency="dotted",
            mapping_dependency="solid",
            composition_dependency="solid",
        ),
        show_hierarchy=True,
        list_variables=False,
        show_root_children=False,
        show_neighbor_children=False,
        show_function_dependencies=False,
    )

    def __init__(
        self,
        digraph: Optional[Dict[str, Any]] = None,
        orientation: Optional[str] = None,
        node_shapes: Optional[Dict[str, str]] = None,
        edge_styles: Optional[Dict[str, str]] = None,
        show_hierarchy: Optional[bool] = None,
        list_variables: Optional[bool] = None,
        show_root_children: Optional[bool] = None,
        show_neighbor_children: Optional[bool] = None,
        show_function_dependencies: Optional[bool] = None,
    ):
        super().__init__(
            digraph=digraph,
            orientation=orientation,
            node_shapes=node_shapes,
            edge_styles=edge_styles,
            show_hierarchy=show_hierarchy,
            list_variables=list_variables,
            show_root_children=show_root_children,
            show_neighbor_children=show_neighbor_children,
            show_function_dependencies=show_function_dependencies,
        )

    @field
    def digraph(self) -> Dict[str, Any]:
        """Options for the :obj:`graphviz.Digraph` object."""

    @field
    def orientation(self) -> str:
        """Orientation of the layout of the graph. One of 'LR' (left-to-right) or
        'TD' (top-down)."""

    @field
    def node_shapes(self) -> Dict[str, str]:
        """Dictionary of node kinds to Graphviz node shapes."""

    @field
    def edge_styles(self) -> Dict[str, str]:
        """Dictionary of edge kind to Graphviz edge styles."""

    @field
    def show_hierarchy(self) -> bool:
        """Whether to draw the nested hierarchical structure."""

    @field
    def list_variables(self) -> bool:
        """Whether to list the variables of goal- and transformation specifications."""

    @field
    def show_root_children(self) -> bool:
        """Whether to display the children of the root component within a functional
        context diagram."""

    @field
    def show_neighbor_children(self) -> bool:
        """Whether to display the children of the neighbor components within a
        functional context diagram."""

    @field
    def show_function_dependencies(self) -> bool:
        """Whether to display dependencies between functions within a traceability
        diagram."""


class Style(Mapping):
    """RaESL plotting style mapping."""

    _defaults = dict(
        diagram=DiagramStyle(),
        ragraph=RaGraphStyle(piemap={"display": "labels", "mode": "relative"}),
    )

    def __init__(
        self,
        diagram: Optional[Union[DiagramStyle, Dict[str, Any]]] = None,
        ragraph: Optional[Union[RaGraphStyle, Dict[str, Any]]] = None,
    ):
        super().__init__(
            diagram=diagram,
            ragraph=ragraph,
        )

    @field
    def diagram(self) -> DiagramStyle:
        """Graphviz diagram style."""

    @field
    def ragraph(self) -> RaGraphStyle:
        """RaGraph style options, used for Multi-Domain matrices."""



"""Line matcher state machines for parameters."""
from typing import TYPE_CHECKING

from raesl.compile.ast.components import VarParam
from raesl.compile.machine_files import typing

if TYPE_CHECKING:
    from raesl.compile.typechecking.ast_builder import AstBuilder


_PARAMETER_HEADER_SPEC = """
parameter_header:
    start initial;
    start -> s1 [PARAMETER_KW] tag=parameter;

    end accept=parameter_header;
    s1 -> end [NL_TK];
    s1 -> end [EOF_TK];
"""


def _process_paramheader(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    token = tags["parameter"][0]
    builder.notify_new_section(token, False)

    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.notify_parameter_section(token)


_PARAMETER_LINE_SPEC = """
parameter_line:
    start initial;
    start -> s1 [NAME] tag=paramname;
    s1 -> start [COMMA_TK];

    s1 -> s2 [IS_KW];
    s2 -> s3 [A_KW];
    s3 -> s4 [NAME] tag=typename;

    end accept=parameter_line;
    s4 -> end [NL_TK];
    s4 -> end [EOF_TK];

    s4 -> s5 [PROPERTY_KW] tag=is_property;
    s5 -> end [NL_TK];
    s5 -> end [EOF_TK];
"""


def _process_paramline(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    paramnames = tags["paramname"]
    typename = tags["typename"][0]
    is_property = "is_property" in tags
    parameters = [VarParam(False, vname, typename, is_property) for vname in paramnames]
    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.add_parameters(parameters)


MACHINES: typing.MachineTripletList = [
    ("PARAMETER_HEADER_MACHINE", _PARAMETER_HEADER_SPEC, _process_paramheader),
    ("PARAMETER_LINE_MACHINE", _PARAMETER_LINE_SPEC, _process_paramline),
]



"""Check that expressions comply with the language requirements."""
from typing import TYPE_CHECKING, Dict, Optional, Set, Tuple, Union

from raesl.compile import diagnostics
from raesl.compile.ast import components
from raesl.compile.ast.exprs import (
    Disjunction,
    ObjectiveComparison,
    RelationComparison,
    Value,
    VariableValue,
)
from raesl.compile.ast.nodes import VarNode
from raesl.compile.typechecking.type_checker import check_type
from raesl.compile.typechecking.utils import resolve_var_param_node

if TYPE_CHECKING:
    from raesl.compile.ast.exprs import DataValue, Expression
    from raesl.compile.ast.types import BaseType
    from raesl.compile.scanner import Token


class ExprChecker:
    """Class for checking expressions.

    Arguments:
        vps: Variables and parameters defined in the expression context.
        diag_store: Storage for found diagnostics.

    Attributes:
        reported_names: Names of variables and variable parts that have been reported
            as error, to avoid duplicate error messages.
    """

    def __init__(
        self,
        vps: Dict[str, Union[components.VarParam]],
        diag_store: diagnostics.DiagnosticStore,
    ):
        self.vps = vps
        self.diag_store = diag_store
        self.reported_names: Set[str] = set()

    def check_expr(self, expr: "Expression") -> bool:
        """Check whether the expression follows all rules of the ESL language.

        Arguments:
            expr: Expression to check.

        Returns:
            Whether the expression is considered to be sufficiently correct to continue
                checking other parts of the construct around the expression.
        """
        is_ok = True
        notdone = [expr]
        while notdone:
            expr = notdone.pop()

            if isinstance(expr, Disjunction):
                # Break the disjunction down to checking its set of child expressions.
                notdone.extend(expr.childs)
                continue

            elif isinstance(expr, RelationComparison):
                if not self.check_relation_comparison(expr):
                    is_ok = False

                continue

            elif isinstance(expr, ObjectiveComparison):
                if not self._check_variable(expr.lhs_var):
                    is_ok = False
                continue

            assert False, "Unexpected expression '{}' found.".format(repr(expr))

        return is_ok

    def check_relation_comparison(self, expr: RelationComparison) -> bool:
        """Check the provided relation comparison.

        Arguments:
            expr: Relation comparison to check.

        Returns:
            Whether the expression is considered to be sufficiently correct.
        """
        # Check both sides.
        left_side = self._check_relation_side(expr.lhs_var)
        right_side = self._check_relation_side(expr.rhs_varval)

        if left_side is None or right_side is None:
            return False

        lhs_pos, lhs_type, lhs_units = left_side
        rhs_pos, rhs_type, rhs_units = right_side

        assert lhs_units is None or len(lhs_units) > 0, "lhs var {} has wrong units '{}'".format(
            expr.lhs_var, lhs_units
        )
        assert rhs_units is None or len(rhs_units) > 0, "rhs varval {} has wrong units '{}'".format(
            expr.rhs_varval, rhs_units
        )

        # Check type compatibility.
        if lhs_type and rhs_type:
            # The lhs and rhs must be compatible in left -> right or in
            # right -> left direction. Additional subtype limits isn't a
            # problem, although it may result in a comparison that can
            # never hold.
            lhs_input = (lhs_pos, lhs_type)
            rhs_input = (rhs_pos, rhs_type)
            diagnostic = check_type(
                supertype=lhs_input, subtype=rhs_input, allow_subtype_limits=True
            )
            if diagnostic:
                diagnostic = check_type(
                    supertype=rhs_input, subtype=lhs_input, allow_subtype_limits=True
                )

                if diagnostic:
                    # A problem in both direction, lhs an rhs cannot be compared.
                    self.diag_store.add(
                        diagnostics.E210(
                            lhs_pos.get_location(),
                            rhs_pos.get_location(),
                            "are not compatible",
                        )
                    )

        # Both sets empty is ok ([-] vs [-])
        if lhs_units or rhs_units:
            # Or if at least one unit is listed anywhere, it must have a common unit.
            lhs_units = set() if lhs_units is None else lhs_units
            rhs_units = set() if rhs_units is None else rhs_units
            if not lhs_units.intersection(rhs_units):
                lhs_loc = lhs_pos.get_location()
                rhs_loc = rhs_pos.get_location()
                self.diag_store.add(
                    diagnostics.E210(lhs_loc, rhs_loc, reason="have no shared unit")
                )

        return True

    def _check_relation_side(
        self, side: "DataValue"
    ) -> Optional[Tuple["Token", Optional["BaseType"], Optional[Set[str]]]]:
        """Check one side in a relation comparison.

        Arguments:
            side: Side to check.

        Returns:
            Information about the checked side. None means an error has been reported,
                otherwise a triplet of token, type, and supported units is returned
                for as far as the side supports each notion.
        """
        if isinstance(side, Value):
            units = side.get_units()
            assert units is None or len(units) > 0, "Value {} has wrong units '{}'".format(
                side, units
            )
            return side.value, None, units

        else:
            assert isinstance(side, VariableValue)
            if not self._check_variable(side):
                return None

            assert isinstance(side.var_node, VarNode)
            typ = side.var_node.the_type
            units = typ.get_units()
            assert units is None or len(units) > 0, "Type {} has wrong units '{}'".format(
                typ, units
            )

            return side.var_tok, typ, units

    def _check_variable(self, var: VariableValue) -> bool:
        """Check that the variable exists in the context.

        Arguments:
            var: Variable to check.

        Returns:
            Whether the variable was found.
        """
        var_node = resolve_var_param_node(
            var.var_tok, self.vps, self.reported_names, self.diag_store
        )
        if var_node is None:
            return False

        var.var_node = var_node
        return True



"""Type checking.

Type compatibility
------------------
Type 'sub_type' is compatible with type 'super_type' if
- Type 'sub_type' is (possibly indirectly) derived from 'super_type'.
- Type 'sub_type' has no additional value constraints in the form of enumerations,
  upper or lower limits, or constants relative to 'super_type'.

The former condition ensures the values are fundamentally compatible. This
condition should always hold, the latter condition ensures that all possible
values of 'super_type' can also be expressed in 'sub_type'. This is particularly
relevant if 'sub_type' may receive data from the element with 'super_type'. If
data flow is in the other direction only, the second condition seems less
relevant.

Note that units are not relevant in this context. If the sub_type is a subtype of
of 'super_type', the former always has all units of the latter.

Relevant code type-classes
--------------------------
Several classes have or use types, or represent data of some type. These classes are

- raesl.compile.ast.types.ElementaryType (type of a single value).
- raesl.compile.ast.types.Compound (type of a bundle of values).
- raesl.compile.ast.nodes.ElementaryVarNode (data of an elementary type).
- raesl.compile.ast.nodes.CompoundVarNode (data of a bundle).
- raesl.compile.ast.nodes.GroupNode (data of a variable group).

where an ElementaryVarNode contains an ElementaryType, a CompoundVarNode
contains a Compound (type), and GroupNode eventually always points at
ElementaryVarNode or CompoundVarNode instances.

The entry_point 'check_type' accepts all the above kinds of objects.
"""
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

from raesl.compile import diagnostics
from raesl.compile.ast.nodes import CompoundVarNode, ElementaryVarNode, GroupNode, Node
from raesl.compile.ast.types import BaseType, Compound, ElementaryType

if TYPE_CHECKING:
    from raesl.compile.scanner import Token

TypeValue = Union[Node, BaseType]


class TypeData:
    """Tuple-like class to keep type data from one side together.

    Arguments:
        name_tok: Initial name of the value in the source, and point of use of the type.
        suffixes: Child texts after the initial name to indicate the subtype value
            being examined. See 'get_name' for a description of the suffixes.
        value: Type or node value being examined.
    """

    def __init__(self, name_tok: "Token", suffixes: List[str], value: TypeValue):
        self.name_tok = name_tok
        self.suffixes = suffixes
        self.value = value

    def get_name(self):
        """Construct a human-readable name of the point being examined.

        It consists of the initial name followed by zero or more suffixes, where
        a suffix can be one of the following
        - ".<childname>" for a child field in a bundle, or
        - "[<1-based index>]" for a variable in a variable group.

        Returns:
            The constructed name.
        """
        return self.name_tok.tok_text + "".join(self.suffixes)

    def drop_to_type(self):
        """For nodes, drop down to its type."""
        if isinstance(self.value, ElementaryVarNode):
            return TypeData(self.name_tok, self.suffixes, self.value.the_type)
        elif isinstance(self.value, CompoundVarNode):
            return TypeData(self.name_tok, self.suffixes, self.value.the_type)
        else:
            return self

    def expand(self):
        """Expand a sequence of values in 'self.value' to a sequence of child TypeData
        instances.
        """
        if isinstance(self.value, Compound):
            childs = []
            for field in self.value.fields:
                child_suffixes = self.suffixes + [".{}".format(field.name.tok_text)]
                child = TypeData(self.name_tok, child_suffixes, field.type)
                childs.append(child)
            return childs

        else:
            assert isinstance(self.value, GroupNode)
            childs = []
            for i, child_node in enumerate(self.value.child_nodes):
                child_suffixes = self.suffixes + ["[{}]".format(i + 1)]
                child = TypeData(self.name_tok, child_suffixes, child_node)
                childs.append(child)
            return childs


def check_type(
    subtype: Tuple["Token", TypeValue],
    supertype: Tuple["Token", TypeValue],
    allow_subtype_limits: bool = False,
) -> Optional[diagnostics.EslDiagnostic]:
    """Like 'check_type_unpacked', except the subtype and super-type data is packed in
    tuples.
    """
    subtok, subval = subtype
    supertok, superval = supertype
    return check_type_unpacked(subtok, subval, supertok, superval, allow_subtype_limits)


def check_type_unpacked(
    sub_nametok: "Token",
    sub_value: TypeValue,
    super_nametok: "Token",
    super_value: TypeValue,
    allow_subtype_limits: bool = False,
) -> Optional[diagnostics.EslDiagnostic]:
    """Check whether sub_value is a subtype of super_value. Iff allow_subtype_limits
    holds, the sub_value may have additional value constraints. Returns None if the
    sub_value is indeed a subtype of super_value, possibly taking additional value
    constraints into account. Otherwise it returns a problem description
    of how it fails.

    Arguments:
        sub_nametok: Text in the input of the sub_value, usually a variable.
        sub_value: Type or node of the sub value.
        super_nametok: Text in the input of the super_value, usually a variable.
        super_value: Type or node of the super value.
        allow_subtype_limits: Whether sub_value may have additional value constraints
            relative to super_value.

    Returns:
        None if sub_value is a subtype of super_value taking allow_subtype_limits
            into account, else a problem description of how it fails to have the
            subtype relation.
    """
    # Do some paranoia checks to ensure checking code does not crash on invalid data.
    assert sub_nametok is not None
    assert super_nametok is not None
    assert isinstance(
        sub_value,
        (ElementaryVarNode, CompoundVarNode, GroupNode, ElementaryType, Compound),
    ), "Weird sub value '{}'".format(sub_value)
    assert isinstance(
        super_value,
        (ElementaryVarNode, CompoundVarNode, GroupNode, ElementaryType, Compound),
    ), "Weird super value '{}'".format(super_value)

    subtype = TypeData(sub_nametok, [], sub_value)
    supertype = TypeData(super_nametok, [], super_value)
    return _check_type(subtype, supertype, allow_subtype_limits)


def _check_type(
    subtype: TypeData, supertype: TypeData, allow_subtype_limits: bool = False
) -> Optional[diagnostics.EslDiagnostic]:
    # Switch subtype and/or supertype from ElementaryVarNode or CompoundVarNode to its
    # type.
    subtype = subtype.drop_to_type()
    supertype = supertype.drop_to_type()

    # Trivially equal?
    if subtype.value is supertype.value:
        return None

    # Handle elementary types
    if isinstance(subtype.value, ElementaryType):
        if not isinstance(supertype.value, ElementaryType):
            # Element does not match with a bundle or variable group.
            return diagnostics.E220(subtype.get_name(), "bundle or variable group")
        else:
            # Compare elementary types
            return _check_elementaries(subtype, supertype, allow_subtype_limits)
    else:
        if isinstance(supertype.value, ElementaryType):
            return diagnostics.E220(supertype.get_name(), "bundle or variable group")

    # Else both are non-elementary types. They must be either Compound or GroupNode

    sub_childs: List[TypeData] = subtype.expand()
    super_childs: List[TypeData] = supertype.expand()

    sub_length = len(sub_childs)
    super_length = len(super_childs)
    if sub_length != super_length:
        return diagnostics.E221(
            "bundle and/or variable group variable",
            sub_length,
            super_length,
            location=subtype.name_tok.get_location(),
            references=[supertype.name_tok.get_location()],
        )

    # Equal length, check pairwise.
    for sub, sup in zip(sub_childs, super_childs):
        diag = _check_type(sub, sup, allow_subtype_limits)
        if diag is not None:
            return diag
    return None


def _check_elementaries(
    subtype: TypeData, supertype: TypeData, allow_subtype_limits: bool = False
) -> Optional[diagnostics.EslDiagnostic]:
    """Check sub/super type relation between two elementary types.

    Returns:
        None if subtype is a subtype of supertype taking allow_subtype_limits into
            account, otherwise it returns a diagnostic describing why the relation does
            not hold.
    """
    sub_val: ElementaryType = subtype.value
    sup_val: ElementaryType = supertype.value
    while True:
        if sup_val is sub_val:
            return None

        # Not the same, sub_val may be a derived type with further limitations.
        # Bail out if not allowed.
        if not allow_subtype_limits and sub_val.intervals is not None:
            return diagnostics.E222(
                subtype.get_name(),
                supertype.get_name(),
                location=subtype.name_tok.get_location(),
                other_loc=supertype.name_tok.get_location(),
            )

        # Walk sub_val up to its direct parent, and check again, unless we run out of
        # parent types.
        if sub_val.parent is None:
            # No more parents!
            return diagnostics.E223(
                subtype.get_name(),
                supertype.get_name(),
                "subtype",
                location=subtype.name_tok.get_location(),
                other_loc=supertype.name_tok.get_location(),
            )

        sub_val = sub_val.parent



"""Utility functions."""
from typing import Dict, List, Optional

from raesl.compile.scanner import Token


def make_loc_names(prefix: str, name: str, count: int) -> Dict[str, str]:
    """Make a dict with 'count' names by combining the prefix, name, and a number."""
    return dict((name + str(num), prefix + name + str(num)) for num in range(1, count + 1))


def _in_range(tok: Token, start_offset: Optional[int], end_offset: Optional[int]) -> bool:
    """Is the given token offset positioned between and not at 'start_offset' and
    'end_offset'? If start_offset or end_offset is None, use -1 and infinity as offsets,
    respectively.

    Returns:
        Whether the provided token is between (not at) the start_offset and end_offset.
    """
    if start_offset is not None and tok.offset <= start_offset:
        return False
    if end_offset is not None and tok.offset >= end_offset:
        return False
    return True


def get_one(tokens: List[Token], start_offset: Optional[int], end_offset: Optional[int]) -> Token:
    """Filter tokens on the provided start and end offsets, and return the only token
    between the positions.
    """
    matches = [tok for tok in tokens if _in_range(tok, start_offset, end_offset)]
    assert len(matches) == 1
    return matches[0]


def get_optional(
    tokens: List[Token], start_offset: Optional[int], end_offset: Optional[int]
) -> Optional[Token]:
    """Filter tokens on the provided start and end offsets, and return the only token
    between the positions.
    """
    matches = [tok for tok in tokens if _in_range(tok, start_offset, end_offset)]
    assert len(matches) < 2
    if matches:
        return matches[0]
    return None



"""Parsing the Elephant Specification Language."""
from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Tuple

from raesl import logger
from raesl.compile import diagnostics, esl_lines
from raesl.compile.scanner import get_token_priority
from raesl.compile.typechecking import ast_builder

if TYPE_CHECKING:
    from raesl.compile.ast.specification import Specification
    from raesl.compile.machine_files.builder import ProcessingStateMachine
    from raesl.compile.scanner import Lexer, Token
    from raesl.compile.state_machine import Location


class LineMachineStepper:
    """Class managing parsing of a line in ESL.

    Arguments:
        machine: Line matching machine to use for recognizing the line.
        lexer: Lexer to use in the matching process.
        current_loc: Current location in the line machine.
        tags: Relevant data extracted from the text line during the parsing process.
        matched_tokens: Tokens used for matching the line thus far.
        dest_loc: New specification location if a match was found.
    """

    def __init__(self, machine: "ProcessingStateMachine", lexer: "Lexer", dest_loc: "Location"):
        self.machine = machine
        self.lexer = lexer
        self.current_loc: Optional["Location"] = machine.initial_loc
        self.tags: Dict[str, List["Token"]] = {}
        self.matched_tokens: List["Token"] = []

        self.dest_loc = dest_loc

    def try_step(self) -> bool:
        """Try to match the next token.

        Returns:
            Whether progress was made.
        """
        if not self.current_loc:
            return False

        match = self.machine.single_step(self.lexer, self.current_loc, self.tags)
        if not match:
            self.current_loc = None
            return False

        self.current_loc = match[0]
        self.matched_tokens.append(match[1])
        return True

    def is_accepting(self) -> bool:
        """Is the machine in an accepting location?"""
        assert self.current_loc
        return self.current_loc.accept is not None

    def get_accept_name(self) -> str:
        """Get the acceptance name associated with the accepting location.

        Returns:
            Name of the accepted 'rule'.
        """
        assert self.current_loc
        assert self.current_loc.accept is not None
        return self.current_loc.accept


def parse_line(
    spec_state: "Location",
    lexer: "Lexer",
    builder: ast_builder.AstBuilder,
    diag_store: diagnostics.DiagnosticStore,
) -> Tuple[Optional["Location"], Optional["Lexer"]]:
    """Parse a text-line in ESL.

    For debugging line selection, set scanner.PARSER_DEBUG to True, which enables
    printing debug information to the std output. For best results, use a *small*
    input specification, output is quite verbose.

    Arguments:
        spec_state: Location in the ESL language state machine.
        lexer: Lexer pointing at the start of the next line to match.
        builder: Class storing extracted parse data.
        diag_store: Storage for reported diagnostics.

    Returns:
        Next state in the ESL state machine unless successfully finished, next lexer to
        use if next step can be performed.
    """
    reachable_locs = collect_locations(spec_state)
    logger.debug(f"** line {lexer.line_num + 1} (1-based) *************************")
    rloc_names = ",".join(rl.name for rl in reachable_locs)
    logger.debug(f"** parse_line({spec_state.name} -> [{rloc_names}])")

    # Skip over empty lines.
    while lexer.find("NL_TK"):
        continue

    # Check for EOF.
    if lexer.find("EOF_TK"):
        if any(loc.accept for loc in reachable_locs):
            return None, lexer  # Reached EOF at an accepting state, success!

        # EOF but not expecting it, report a problem.
        diag_store.add(diagnostics.E100(lexer.get_location()))
        return spec_state, None

    # Found a line of text. Find a match.
    #
    # Create line steppers for all possible matches.
    steppers = []
    for spec_loc in reachable_locs:
        for edge in spec_loc.out_edges:
            if edge.tok_type == "epsilon":
                continue

            machine = esl_lines.get_line_machine(edge.tok_type)
            logger.debug(f"** Add '{machine.name}' line machine.")
            steppers.append(LineMachineStepper(machine, lexer.copy(), edge.dest))

    assert steppers  # There should be at least one stepper.

    # Take steps, silently dropping steppers that don't match, until all don't match or
    # at least one stepper is accepting.
    acceptors = []
    while steppers:
        prev_steppers = steppers
        steppers = []
        logger.debug("** -----")
        steppers_text = ", ".join(s.machine.name for s in prev_steppers)
        logger.debug(f"** Steppers remaining: {steppers_text}")
        for stepper in prev_steppers:
            if not stepper.try_step():
                logger.debug(f"** Stepper {stepper.machine.name} didn't match.")
                continue  # Next token didn't match, drop the stepper silently.
            if stepper.is_accepting():
                # Stepper has reached the end, continue with the others.
                logger.debug(f"** Stepper {stepper.machine.name} has reached the end.")
                acceptors.append(stepper)
                continue

            logger.debug(f"** Stepper {stepper.machine.name} matched.")

            steppers.append(stepper)

    # All steppers either failed or finished in an accepting location.
    best_acceptor: Optional[LineMachineStepper]
    if acceptors:
        if len(acceptors) == 1:
            # Life is simple, pick the one and only match.
            best_acceptor = acceptors[0]
        else:
            # Multiple lines match. Filter on the most specific match.
            best_match = None
            best_acceptor = None
            for acceptor in acceptors:
                match_prio = [get_token_priority(tok.tok_type) for tok in acceptor.matched_tokens]
                if best_match is None or match_prio < best_match:
                    best_match = match_prio
                    best_acceptor = acceptor
                elif best_acceptor is not None and match_prio == best_match:
                    ambi_acceptors = (
                        best_acceptor.get_accept_name(),
                        acceptor.get_accept_name(),
                    )
                    best_acceptor = None  # Ambiguous best match (until now)

            # Sanity check and raise an Exception if it fails.
            if best_acceptor is None:
                diag_store.add(diagnostics.E101(ambi_acceptors, location=lexer.get_location()))

        # Store information of the line into the ast builder instance.
        assert best_acceptor is not None
        processing_func = best_acceptor.machine.processing_func
        if processing_func:
            tags = best_acceptor.tags
            accept = best_acceptor.get_accept_name()
            processing_func(tags, accept, builder)

        # Line done.
        return best_acceptor.dest_loc, best_acceptor.lexer

    # No acceptors, thus they all failed to match. Find a stepper that got the furthest.
    # mypy fails on lambdas
    # best_stepper = max(prev_steppers, key=lambda s: s.lexer.get_linecol())
    def linecol_value(line_stepper: LineMachineStepper) -> Tuple[int, int]:
        return line_stepper.lexer.get_linecol()

    # Report a syntax error
    best_stepper = max(prev_steppers, key=linecol_value)
    diag_store.add(diagnostics.E102(location=best_stepper.lexer.get_location()))

    return spec_state, None


def parse_lexer(
    lexer: "Lexer",
    diag_store: Optional[diagnostics.DiagnosticStore],
    builder: Optional[ast_builder.AstBuilder],
    doc_comments: Optional[List["Token"]],
) -> Tuple[diagnostics.DiagnosticStore, ast_builder.AstBuilder, List["Token"], bool]:
    """Parse an ESL lexer, storing collected information in the builder.

    Arguments:
        lexer: Lexer pointing at the start of the specification text.
        diag_store: Diagnostic store if one already has been created.
        builder: Builder if one already has been created.
        doc_comments: Doc comments if any have been found yet.

    Returns:
        Diagnostic store instance.
        Builder instance.
        Found doc comments.
        Whether there has been an error.
    """
    diag_store = diagnostics.DiagnosticStore() if diag_store is None else diag_store
    builder = ast_builder.AstBuilder(diag_store) if builder is None else builder
    doc_comments = [] if doc_comments is None else doc_comments

    assert esl_lines.ESL_MACHINE.initial_loc is not None
    spec_state: Location = esl_lines.ESL_MACHINE.initial_loc
    while True:
        new_spec_state, new_lexer = parse_line(spec_state, lexer, builder, diag_store)
        if new_lexer is None:
            # Something bad happened, message should be in problem storage.
            doc_comments.extend(lexer.doc_comments)
            return diag_store, builder, doc_comments, True

        if new_spec_state is None:
            # EOF reached in an accepting state, done!
            # Check the collected data and construct an AST, return the found
            # diagnostics and the created specification if possible.
            doc_comments.extend(lexer.doc_comments)
            return diag_store, builder, doc_comments, False

        # Else, matched one line, do the next.
        spec_state = new_spec_state
        lexer = new_lexer


def parse_spec(
    lexers: Iterable["Lexer"], diag_store: Optional[diagnostics.DiagnosticStore] = None
) -> Tuple[diagnostics.DiagnosticStore, Optional["Specification"]]:
    """Parse an ESL specification, storing collected information in the builder.

    Arguments:
        lexers: Lexers pointing at the start of their respective texts (e.g. per-file).

    Returns:
        The found diagnostics, and if successful, the type-checked output.
    """
    diag_store = diagnostics.DiagnosticStore() if diag_store is None else diag_store
    builder = ast_builder.AstBuilder(diag_store)
    doc_comments: List["Token"] = []

    for lexer in lexers:
        diag_store, builder, doc_comments, error = parse_lexer(
            lexer, diag_store, builder, doc_comments
        )
        if error:
            return diag_store, None

    spec = builder.finish_parse(doc_comments)
    return diag_store, spec


def collect_locations(spec_state: "Location") -> List["Location"]:
    """Collect the set reachable states from 'spec_state' without taking only 'epsilon'
    transitions.
    """
    reachables = [spec_state]
    notdone = [spec_state]
    while notdone:
        spec_loc = notdone.pop()
        for edge in spec_loc.out_edges:
            if edge.tok_type == "epsilon":
                new_loc = edge.dest
                if (
                    new_loc not in reachables
                ):  # Assuming a msall number of location in 'reachables'.
                    reachables.append(new_loc)
                    notdone.append(new_loc)
    return reachables



"""Doc generation utility functions."""

from typing import List

from ragraph.graph import Graph
from ragraph.node import Node


def get_component_goals(
    component: Node, graph: Graph, constraint: bool = True, inherited: bool = True
) -> List[Node]:
    """Get relevant goal requirements or constraints for a component."""
    form = "constraint" if constraint else "requirement"
    ancestors = set([a.name for a in component.ancestors]) if inherited else set()
    goals = [
        n
        for n in graph.nodes
        if n.kind == "function_spec"
        and n.annotations.esl_info.get("sub_kind") == "goal"
        and n.annotations.esl_info.get("form") == form
        and (
            n.annotations.esl_info["body"].get("active") in ancestors
            if inherited
            else n.annotations.esl_info["body"].get("active") == component.name
        )
        and [e for e in graph.edges_between(component, n) if e.kind == "mapping_dependency"]
    ]
    return goals


def get_component_transformations(
    component: Node, graph: Graph, constraint: bool = True
) -> List[Node]:
    """Get relevant transformation requirements or constraints for a component."""
    form = "constraint" if constraint else "requirement"
    transformations = [
        n
        for n in graph.nodes
        if n.kind == "function_spec"
        and n.annotations.esl_info.get("sub_kind") == "transformation"
        and n.annotations.esl_info.get("form") == form
        and n.annotations.esl_info["body"].get("active") == component.name
    ]
    return transformations


def get_component_behaviors(component: Node, graph: Graph, constraint: bool = True) -> List[Node]:
    """Get relevant behavior requirements or constraints for a component."""
    form = "constraint" if constraint else "requirement"
    return [
        b
        for b in graph.nodes
        if b.kind == "behavior_spec"
        and b.annotations.esl_info.get("form") == form
        and graph[component.name, b.name]
    ]


def get_component_designs(component: Node, graph: Graph, constraint: bool = True) -> List[Node]:
    """Get relevant design requirements or constraints for a component."""
    form = "constraint" if constraint else "requirement"
    return [
        d
        for d in graph.nodes
        if d.kind == "design_spec"
        and d.annotations.esl_info.get("form") == form
        and graph[component.name, d.name]
    ]


def get_global_designs(graph: Graph, constraint: bool = True) -> List[Node]:
    """Get globally relevant design requirments or constraints."""
    form = "constraint" if constraint else "requirement"
    dc_dict = {}
    for e in graph.edges:
        if e.source.kind != "component":
            continue
        dc_dict[e.target.name] = e.source.name

    drs = [
        d
        for d in graph.nodes
        if d.kind == "design_spec"
        and d.annotations.esl_info.get("form") == form
        and d.name not in dc_dict
    ]

    return drs


def get_component_needs(component: Node, graph: Graph) -> List[Node]:
    """Get relevant needs for a component."""
    subjects = set([n.name for n in graph.targets_of(component) if n.kind != "component"])
    subjects.add(component.name)
    return [
        n for n in graph.nodes if n.kind == "need" and n.annotations.esl_info["subject"] in subjects
    ]


def get_global_needs(graph: Graph) -> List[Node]:
    """Get globally relevant needs."""
    sc_dict = {}
    for e in graph.edges:
        if e.source.kind != "component":
            continue
        sc_dict[e.target.name] = e.source.name

    all_needs = graph.get_nodes_by_kind("need")

    def get_subject(need):
        return need.annotations.esl_info["subject"]

    return [
        need
        for need in all_needs
        if get_subject(need) not in sc_dict and graph[get_subject(need)].kind != "component"
    ]


def get_component_relations(component: Node, graph: Graph) -> List[Node]:
    """Get relevant relations for a component."""
    return [n for n in graph.nodes if n.kind == "relation_spec" and graph[component.name, n.name]]


def get_component_properties(component: Node, graph: Graph) -> List[Node]:
    """Get relevant properties for a component."""
    return [graph[prop] for prop in component.annotations.esl_info.get("property_variables", [])]
